import subprocess
import atexit
import os
import sys
from subprocess import PIPE, TimeoutExpired, Popen
from typing import Optional, Tuple, List
from loguru import logger

""" Author: Ondrej Borysek, License: MIT, Last update: 2021-04-19"""

SIGKILL_STATUS_CODE = 137 - 128  # -9
SIGTERM_STATUS_CODE = 143 - 128  # -15


class ProcessHandler:
    """
    This simple wrapper for subprocess library enables easier Input/Output testing of other programs.
    No pip packages or additional linux packages are required.
    Should work on Linux, Windows, and hopefully also on MacOS.
    """

    def __init__(self, timeout: float) -> None:
        self.last_child: Optional[Popen[bytes]] = None
        self.__try_to_be_nice()
        self.timeout: float = timeout
        self.sigterm_grace_period = 0.2  # second
        atexit.register(self.__kill_children)
        # Using prctl would be safer, but less portable: https://pythonhosted.org/python-prctl/#prctl.set_pdeathsig

    def __try_to_be_nice(self) -> None:
        if getattr(os, "nice", None):
            try:
                os.nice(10)
            except Exception as e:
                logger.warning(f"System suports 'nice', but the attempt failed with {e}")  # This happens with Ubuntu snaps.

    def __kill_children(self) -> None:
        if self.last_child is None or self.last_child.poll() is not None:
            return

        logger.info("trying to terminate")
        self.last_child.terminate()  # Ask the child to exit peacefully.

        try:
            self.last_child.wait(timeout=self.sigterm_grace_period)
        except TimeoutExpired:
            logger.info("child refused to terminate; trying SIGKILL if available")
            self.last_child.kill()  # Kill anyone still standing, just like Anakin did.
            # self.last_child.wait()  # Waiting would be safe on Linux, but on Windows .kill from Python
            # is just .terminate

        if self.last_child.poll():
            try:
                _, _ = self.last_child.communicate(
                    timeout=0.2
                )  # this can garbage collect the process
            except TimeoutExpired:
                logger.warning(
                    "the process refused to die, you might need to kill the zombi manually"
                )
                ProcessHandler.linux_print_processes()

    def __start_process(
        self, user_command: List[str], input_str: Optional[str] = None
    ) -> Tuple[int, str, str]:
        # Subprocess will be with the same niceness as the main program.
        # Warning: Do NOT use shell=True. That would only kill the shell, not the C program.
        proc = subprocess.Popen(user_command, shell=False, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        self.last_child = proc

        return_code = SIGKILL_STATUS_CODE

        try:
            outb, errb = proc.communicate(
                input=input_str.encode("utf8") if input_str else None, timeout=self.timeout
            )
            return_code = proc.returncode
        except TimeoutExpired:
            logger.warning("timeout, trying to kill")
            proc.kill()
            outb, errb = proc.communicate()
            # proc.returncode will be ignored

        self.__kill_children()
        return return_code, outb.decode(), errb.decode()

    @staticmethod
    def run(
        command: List[str],
        input_str: Optional[str] = None,
        timeout: float = 5,  # seconds
    ) -> Tuple[int, str, str]:
        if command is None:
            return -1, "", "No command provided"

        ph = ProcessHandler(timeout=timeout)
        return_code, outs, errs = ph.__start_process(command, input_str=input_str)
        logger.trace(
            "Command: {command}\n"
            "Input:\n---\n{input_str}\n---\n"
            "Return code: {return_code}\n"
            "Output:\n---\n{output_str}\n---\n"
            "Error:\n---\n{error_str}\n---\n",
            command=command,
            input_str=input_str,
            return_code=return_code,
            output_str=ProcessHandler.prettyfi_the_output(outs),
            error_str=ProcessHandler.prettyfi_the_output(errs),
        )
        return return_code, outs, errs

    @staticmethod
    def prettyfi_the_output(a: str) -> str:
        return a.replace("\\r\\n", "\n").replace("\\n", "\n")

    @staticmethod
    def linux_print_processes() -> None:
        if not sys.platform.startswith("linux"):
            return
        logger.warning(
            "please check that you do not see any zombie processes from running this script, "
            "they would have nice value of 19. You can use command 'ps a -o pid,ni,time,cmd'"
        )

    @staticmethod
    def is_status_code_by_timeout(status_code: int) -> bool:
        return status_code in [SIGTERM_STATUS_CODE, SIGKILL_STATUS_CODE]



from astroid import nodes  # type: ignore
from typing import TYPE_CHECKING

from pylint.checkers import BaseChecker  # type: ignore
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.checkers.utils import get_statements_count
from edulint.linting.analyses.cfg.utils import successors_from_loc
from edulint.linting.analyses.var_events import VarEventType


class NoGlobalVars(BaseChecker):
    name = "no-global-variables"
    msgs = {
        "R6401": (
            "Do not use global variables; you use %s, modifying it for example at line %i.",
            "no-global-vars",
            "Emitted when the code uses global variables.",
        ),
    }

    def visit_module(self, node: nodes.Module):
        if len(node.body) == 0:
            return

        toplevel_vars = {}
        for loc in successors_from_loc(node.body[0].cfg_loc, include_start=True):
            toplevel = loc.node
            if isinstance(toplevel, (nodes.Import, nodes.ImportFrom)):
                continue

            for var, event in toplevel.cfg_loc.var_events.all():
                if var.scope == node and event.type in (VarEventType.ASSIGN, VarEventType.REASSIGN):
                    toplevel_vars[var] = event

        if len(toplevel_vars) == 0:
            return

        for loc in successors_from_loc(
            node.cfg_loc, include_start=True, explore_functions=True, explore_classes=True
        ):
            for var, events in loc.var_events.items():
                if var not in toplevel_vars:
                    continue
                for event in events:
                    if event.type != VarEventType.READ and event.node.parent.scope() != node:
                        self.add_message(
                            "no-global-vars",
                            node=toplevel_vars[var].node,
                            args=(var.name, event.node.lineno),
                        )
                        toplevel_vars.pop(var)
                        break


class LongCodeChecker(BaseChecker):
    name = "long-code"
    msgs = {
        "R6701": (
            "Too much code outside of functions or classes (%d which is over %d statements).",
            "long-script",
            "Emitted when there are too many lines of code on the top level that are not import or function or class "
            "definition.",
        ),
        "R6702": (
            "Function '%s' is too long (%d which is over %d statements).",
            "long-function",
            "Emitted when there are too many statements inside a function definition.",
        ),
    }

    @only_required_for_messages("long-script")
    def visit_module(self, node: nodes.Module):
        MAX_SCRIPT = 20

        count = get_statements_count(node, include_defs=False, include_name_main=False)
        if count > MAX_SCRIPT:
            self.add_message("long-script", node=node, args=(count, MAX_SCRIPT))

    @only_required_for_messages("long-function")
    def visit_functiondef(self, node: nodes.FunctionDef):
        MAX_FUNC = 20

        count = get_statements_count(node.body, include_defs=False, include_name_main=False)
        if count > MAX_FUNC:
            self.add_message("long-function", node=node, args=(node.name, count, MAX_FUNC))


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(NoGlobalVars(linter))
    linter.register_checker(LongCodeChecker(linter))



from typing import List, Union, Set, Optional, Dict, Tuple
from collections import defaultdict

from astroid import nodes

from edulint.linting.analyses.cfg.graph import CFGLoc, CFGBlock
from edulint.linting.analyses.cfg.utils import (
    successor_blocks_from_locs,
    predecessors_from_loc,
    syntactic_children_locs,
    syntactic_children_locs_from,
    get_cfg_loc,
)
from edulint.linting.analyses.var_events import VarEventType, Variable, VarEvent, strip_to_name


MODIFYING_EVENTS = (VarEventType.ASSIGN, VarEventType.REASSIGN, VarEventType.MODIFY)
KILLING_EVENTS = (
    VarEventType.ASSIGN,
    VarEventType.REASSIGN,
    VarEventType.DELETE,
    VarEventType.MODIFY,
)
GENERATING_EVENTS = (VarEventType.READ, VarEventType.MODIFY)


### collectors


def collect_gens_kill(
    block: CFGBlock,
) -> Tuple[Dict[Variable, List[VarEvent]], Dict[Variable, List[VarEvent]]]:
    gens, kill = defaultdict(list), defaultdict(list)
    for loc in block.locs:
        for var, event in loc.var_events.all():
            if event.type in GENERATING_EVENTS:
                if var not in kill:
                    gens[var].append(event)
                else:
                    assert len(kill[var]) == 1
                    kill_event = kill[var][0]

                    if kill_event not in event.definitions:
                        event.definitions.append(kill_event)

                    if event not in kill_event.uses:
                        kill_event.uses.append(event)

            if event.type in KILLING_EVENTS:
                kill[var] = [event]

    return gens, kill


def kills_from_parents(
    block: CFGBlock,
    kills: Dict[CFGBlock, Dict[Variable, List[Tuple[CFGLoc, VarEvent]]]],
    parent_scope_kills: Dict[Variable, List[Tuple[CFGLoc, VarEvent]]],
) -> Dict[Variable, List[Tuple[CFGLoc, VarEvent]]]:
    if len(block.predecessors) == 0:
        return parent_scope_kills

    result = defaultdict(list)
    for edge in block.predecessors:
        parent = edge.source
        if not parent.reachable:
            continue
        for var, es in kills[parent].items():
            result[var].extend(es)
    return result


def collect_reaching_definitions(
    node: Union[nodes.Module, nodes.Arguments],
    parent_scope_kills: Dict[Variable, List[Tuple[CFGLoc, nodes.NodeNG]]] = None,
) -> None:
    parent_scope_kills = parent_scope_kills if parent_scope_kills is not None else {}

    original_gens_kills = {}
    blocks = []
    for block, start, end in successor_blocks_from_locs([node.cfg_loc], include_start=True):
        assert start == 0 and end == len(block.locs)
        original_gens_kills[block] = collect_gens_kill(block)

        blocks.append(block)

    computed_kills = {
        block: defaultdict(list, {var: k.copy() for var, k in kills.items()})
        for block, (_gens, kills) in original_gens_kills.items()
    }

    changed = True
    while changed:
        changed = False

        for block in blocks:
            parent_kills = kills_from_parents(block, computed_kills, parent_scope_kills)
            gens, original_kill = original_gens_kills[block]
            computed_kill = computed_kills[block]

            for var, es in parent_kills.items():
                # this block does not kill the parent's value
                if var not in original_kill:
                    for kill_event in es:
                        # if the kill was not added already
                        if kill_event not in computed_kill[var]:
                            computed_kill[var].append(kill_event)
                            changed = True

                for gen_event in gens.get(var, []):
                    for kill_event in es:
                        if kill_event not in gen_event.definitions:
                            gen_event.definitions.append(kill_event)
                        if gen_event not in kill_event.uses:
                            kill_event.uses.append(gen_event)

    for block in blocks:
        for loc in block.locs:
            if isinstance(loc.node, nodes.FunctionDef):
                collect_reaching_definitions(loc.node.args, computed_kills[block])
            elif isinstance(loc.node, nodes.ClassDef):
                for class_node in loc.node.body:
                    if isinstance(class_node, nodes.FunctionDef):
                        collect_reaching_definitions(class_node.args, computed_kills[block])


### var getters


def vars_in(
    node: Union[nodes.NodeNG, List[nodes.NodeNG]], event_types: Optional[Set[VarEventType]] = None
) -> Dict[Variable, List[nodes.NodeNG]]:
    result = {}

    if isinstance(node, nodes.NodeNG):
        loc_node = get_cfg_loc(node).node
    else:
        loc_node = node

    for loc in syntactic_children_locs(loc_node):
        for var, event in loc.var_events.all():
            if event_types is None or event.type in event_types:
                if loc_node == node or (event.node == node or node in event.node.node_ancestors()):
                    result[var] = event.node
    return result


def name_to_var(name: str, node: nodes.NodeNG):
    start = get_cfg_loc(node)
    for loc in syntactic_children_locs_from(start, node):
        for var, events in loc.var_events.items():
            event_scope = events[0].node.scope()
            if var.name == name and event_scope == node.scope():
                return var
    return None


def node_to_var(node: nodes.NodeNG):
    loc = get_cfg_loc(node)
    for var, event in loc.var_events.all():
        if event.node == node:
            return var
    return None


def _get_matching_superpart(var_node: nodes.NodeNG, super_node: nodes.NodeNG):
    stripped = strip_to_name(super_node)
    while stripped != super_node:
        stripped = stripped.parent
        var_node = var_node.parent

        if not isinstance(var_node, type(stripped)):
            return None

        if isinstance(stripped, (nodes.Attribute, nodes.AssignAttr, nodes.DelAttr)):
            if stripped.attrname != var_node.attrname:
                return None

        elif isinstance(stripped, nodes.Subscript):
            stripped_vars = set(vars_in(stripped.slice).keys())
            node_vars = set(vars_in(var_node.slice).keys())
            stripped_events = get_events_by_var(stripped_vars, [stripped.slice])
            node_events = get_events_by_var(node_vars, [var_node.slice])
            if list(stripped_events.keys()) != list(node_events.keys()):
                return None

            for var in stripped_events.keys():
                if [e.definitions for e in stripped_events[var]] != [
                    e.definitions for e in node_events[var]
                ]:
                    return None
        else:
            assert False, "unreachable, but " + str(type(stripped))
    return var_node


def filter_events_for(node: nodes.Attribute, events: List[VarEvent]) -> List[VarEvent]:
    if isinstance(node, (nodes.Name, nodes.AssignName)):
        return events

    result = []
    for event in events:
        superpart = _get_matching_superpart(event.node, node)
        if superpart is not None:
            result.append(VarEvent(event.var, superpart, event.type, None, None))
    return result


### event checkers


def is_changed_between(var: Variable, before_loc: CFGLoc, after_locs: List[CFGLoc]) -> bool:
    for after_loc in after_locs:
        for loc in predecessors_from_loc(
            after_loc, stop_on_loc=lambda loc: loc == before_loc, include_start=True
        ):
            for event in loc.var_events[var]:
                if event.type != VarEventType.READ:
                    return True
    return False


def get_events_for(
    vars: List[Variable],
    nodes: List[nodes.NodeNG],
    event_types: Optional[List[VarEventType]] = None,
):
    if len(nodes) == 0:
        return

    loc_nodes = [get_cfg_loc(n).node for n in nodes]
    check_ancestors = all(n == ln for n, ln in zip(nodes, loc_nodes))
    for loc in syntactic_children_locs(loc_nodes, explore_functions=True, explore_classes=True):
        for var in vars:
            for event in loc.var_events.for_var(var):
                if (event_types is None or event.type in event_types) and (
                    not check_ancestors
                    or any(event.node == n or n in event.node.node_ancestors() for n in nodes)
                ):
                    yield event


def get_events_by_var(
    vars: List[Variable],
    nodes: List[nodes.NodeNG],
    event_types: Optional[List[VarEventType]] = None,
):
    result = defaultdict(list)
    for event in get_events_for(vars, nodes, event_types):
        result[event.var].append(event)
    return result


def modified_in(vars: List[Variable], nodes: List[nodes.NodeNG]) -> bool:
    # just check whether there is such an event
    for _ in get_events_for(vars, nodes, MODIFYING_EVENTS):
        return True
    return False


def get_vars_defined_before(core) -> Dict[Variable, Set[nodes.NodeNG]]:
    # FIXME counts all read variables, including avars
    result = defaultdict(set)
    for i in range(len(core[0].sub_locs)):
        core_sub_locs = [c.sub_locs[i] for c in core]
        children_locs = set(
            syntactic_children_locs_from(core_sub_locs[0], [loc.node for loc in core_sub_locs])
        )
        for loc in children_locs:
            for var, event in loc.var_events.all():
                # if event.type not in GENERATING_EVENTS:
                #     continue
                for def_event in event.definitions:
                    if get_cfg_loc(def_event.node) not in children_locs:
                        result[var].add(def_event.node)
    return result


def get_vars_used_after(core) -> Dict[Variable, Set[nodes.NodeNG]]:
    """DANGER: variables used in core are not returned, even if core is in a loop"""
    result = defaultdict(set)
    for i in range(len(core[0].sub_locs)):
        core_sub_locs = [c.sub_locs[i] for c in core]
        children_locs = set(
            syntactic_children_locs_from(core_sub_locs[0], [loc.node for loc in core_sub_locs])
        )
        for loc in children_locs:
            for var, event in loc.var_events.all():
                # if event.type not in MODIFYING_EVENTS:
                #     continue
                for use_event in event.uses:
                    if get_cfg_loc(use_event.node) not in children_locs:
                        result[var].add(use_event.node)
    return result


def get_control_statements(core):
    result = []
    first_loc = get_cfg_loc(core[0] if isinstance(core, list) else core)
    for loc in syntactic_children_locs_from(first_loc, core):
        if isinstance(loc.node, (nodes.Return, nodes.Break, nodes.Continue)):
            result.append(loc.node)
    return result



from astroid import nodes  # type: ignore
from typing import TYPE_CHECKING, Optional, Tuple, List, Union, Set, Any, Dict

from pylint.checkers import BaseChecker  # type: ignore
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.checkers.utils import (
    is_parents_elif,
    BaseVisitor,
    is_any_assign,
    get_lines_between,
    is_main_block,
    is_block_comment,
)


class InvalidExpression(Exception):
    pass


class DuplicateExprVisitor(BaseVisitor[None]):
    EXPR_FUNCTIONS = {
        "abs",
        "max",
        "min",
        "round",
        "sqrt",
        "len",
        "all",
        "any",
        "sum",
        "map",
        "filter",
        "sorted",
        "reversed",
        "int",
        "str",
        "ord",
        "chr",
    }

    def __init__(self) -> None:
        super().__init__()
        self.long_expressions: Dict[str, List[nodes.NodeNG]] = {}

    @classmethod
    def _compute_complexity(cls, node: nodes.NodeNG) -> int:
        fun = cls._compute_complexity

        if isinstance(node, (nodes.Attribute, nodes.Subscript)) and is_any_assign(node.parent):
            raise InvalidExpression

        if isinstance(node, nodes.Call) and node.func.as_string() not in cls.EXPR_FUNCTIONS:
            raise InvalidExpression

        if isinstance(node, nodes.BinOp):
            return 2 + fun(node.left) + fun(node.right)

        if isinstance(node, nodes.BoolOp):
            return len(node.values) - 1 + sum(map(fun, node.values))

        if isinstance(node, nodes.Compare):
            return fun(node.left) + sum(1 + fun(op) for _, op in node.ops)

        if isinstance(node, (nodes.Name, nodes.Const)):
            return 1

        if isinstance(node, nodes.Attribute):
            return 1 + sum(map(fun, node.get_children()))

        return sum(map(fun, node.get_children()))

    def _process_expr(self, node: Optional[nodes.NodeNG]) -> None:
        if node is None:
            return

        try:
            complexity = DuplicateExprVisitor._compute_complexity(node)
        except InvalidExpression:
            self.visit_many(node.get_children())
            return

        if complexity >= 8:  # TODO extract into parameter
            others = self.long_expressions.get(node.as_string(), [])
            others.append(node)
            self.long_expressions[node.as_string()] = others

            self.visit_many(node.get_children())

    def visit_assert(self, node: nodes.Assert) -> None:
        return

    def visit_attribute(self, node: nodes.Attribute) -> None:
        self._process_expr(node)

    def visit_binop(self, node: nodes.BinOp) -> None:
        self._process_expr(node)

    def visit_boolop(self, node: nodes.BoolOp) -> None:
        self._process_expr(node)

    def visit_call(self, node: nodes.Call) -> None:
        self._process_expr(node)

    def visit_compare(self, node: nodes.Compare) -> None:
        self._process_expr(node)

    def visit_dict(self, node: nodes.Dict) -> None:
        self._process_expr(node)

    def visit_dictcomp(self, node: nodes.DictComp) -> None:
        self._process_expr(node)

    def visit_ifexp(self, node: nodes.IfExp) -> None:
        self._process_expr(node)

    def visit_joinedstr(self, node: nodes.JoinedStr) -> None:
        self._process_expr(node)

    def visit_lambda(self, node: nodes.Lambda) -> None:
        self._process_expr(node)

    def visit_list(self, node: nodes.List) -> None:
        self._process_expr(node)

    def visit_listcomp(self, node: nodes.ListComp) -> None:
        self._process_expr(node)

    # def visit_name(self, node: nodes.Name) -> None:
    #     self._process_expr(node)

    def visit_set(self, node: nodes.Set) -> None:
        self._process_expr(node)

    def visit_setcomp(self, node: nodes.SetComp) -> None:
        self._process_expr(node)

    def visit_starred(self, node: nodes.Starred) -> None:
        self._process_expr(node)

    def visit_subscript(self, node: nodes.Subscript) -> None:
        self._process_expr(node)

    def visit_tuple(self, node: nodes.Tuple) -> None:
        self._process_expr(node)

    def visit_unaryop(self, node: nodes.UnaryOp) -> None:
        self._process_expr(node)


class CollectBlocksVisitor(BaseVisitor[None]):
    def __init__(self) -> None:
        self.blocks: List[List[nodes.NodeNG]] = []

    def visit_if(self, node: nodes.If) -> None:
        self.blocks.append(node.body)

        if len(node.orelse) > 0 and not node.has_elif_block():
            self.blocks.append(node.orelse)

        self.visit_many(node.body)
        self.visit_many(node.orelse)

    def visit_loop(self, node: Union[nodes.For, nodes.While]) -> None:
        self.blocks.append(node.body)

        if len(node.orelse) > 0:
            self.blocks.append(node.orelse)

        self.visit_many(node.body)
        self.visit_many(node.orelse)

    def visit_for(self, node: nodes.For) -> None:
        self.visit_loop(node)

    def visit_while(self, node: nodes.While) -> None:
        self.visit_loop(node)

    def visit_with(self, node: nodes.With) -> None:
        self.blocks.append(node.body)
        self.visit_many(node.body)

    def visit_try(self, node: nodes.Try) -> None:
        self.blocks.append(node.body)

        if len(node.handlers) > 0:
            self.blocks.extend(h.body for h in node.handlers)

        if len(node.orelse) > 0:
            self.blocks.append(node.orelse)

        if len(node.finalbody) > 0:
            self.blocks.append(node.finalbody)

        self.visit_many(node.body)
        self.visit_many([stmt for h in node.handlers for stmt in h.body])
        self.visit_many(node.orelse)
        self.visit_many(node.finalbody)

    def visit_functiondef(self, node: nodes.FunctionDef) -> None:
        self.blocks.append(node.body)

        self.visit_many(node.body)

    def visit_module(self, node: nodes.Module) -> None:
        self.blocks.append(
            [
                b
                for b in node.body
                if b
                and not isinstance(b, (nodes.FunctionDef, nodes.ClassDef))
                and not is_main_block(b)
            ]
        )

        self.visit_many(node.body)


class NoDuplicateCode(BaseChecker):  # type: ignore
    name = "old-no-duplicate-code"
    msgs = {
        "R6595": (
            "Identical code inside all if's branches, move %d lines %s the if.",
            "duplicate-if-branches",
            "Emitted when identical code starts or ends all branches of an if statement.",
        ),
        "R6596": (
            "Identical code inside %d consecutive ifs, join their conditions using 'or'.",
            "duplicate-seq-ifs",
            "Emitted when several consecutive if statements have identical bodies and thus can be "
            "joined by or in their conditions.",
        ),
        "R6597": (
            "A complex expression '%s' used repeatedly (on lines %s). Extract it to a local variable or "
            "create a helper function.",
            "duplicate-exprs",
            "Emitted when an overly complex expression is used multiple times.",
        ),
        "R6598": (
            "Duplicate blocks starting on lines %s. Extract the code to a helper function.",
            "duplicate-blocks",
            "Emitted when there are duplicate blocks of code as a body of an if/elif/else/for/while/with/try-except "
            "block.",
        ),
        "R6599": (
            "Duplicate sequence of %d repetitions of %d lines of code. Use a loop to avoid this.",
            "duplicate-sequence",
            "Emitted when there is a sequence of similar sub-blocks inside a block that can be replaced by a loop.",
        ),
    }

    @only_required_for_messages("duplicate-if-branches", "duplicate-seq-ifs")
    def visit_if(self, node: nodes.If) -> None:
        self.duplicate_if_branches(node)
        self.duplicate_seq_ifs(node)

    def duplicate_if_branches(self, node: nodes.If) -> None:
        def extract_branch_bodies(node: nodes.If) -> Optional[List[nodes.NodeNG]]:
            branches = [node.body]
            current = node
            while current.has_elif_block():
                elif_ = current.orelse[0]
                if not elif_.orelse:
                    return None

                branches.append(elif_.body)
                current = elif_
            branches.append(current.orelse)
            return branches

        def get_stmts_difference(branches: List[nodes.NodeNG], forward: bool) -> int:
            reference = branches[0]
            compare = branches[1:]
            for i in range(min(map(len, branches))):
                for branch in compare:
                    index = i if forward else -i - 1
                    if reference[index].as_string() != branch[index].as_string():
                        return i
            return i + 1

        def add_message(
            branches: List[nodes.NodeNG],
            stmts_difference: int,
            defect_node: nodes.NodeNG,
            forward: bool = True,
        ) -> None:
            reference = branches[0]
            first = reference[0 if forward else -stmts_difference]
            last = reference[stmts_difference - 1 if forward else -1]
            lines_difference = get_lines_between(first, last, including_last=True)

            self.add_message(
                "duplicate-if-branches",
                node=defect_node,
                args=(lines_difference, "before" if forward else "after"),
            )

        if not node.orelse or is_parents_elif(node):
            return

        branches = extract_branch_bodies(node)
        if branches is None:
            return

        same_prefix_len = get_stmts_difference(branches, forward=True)
        if same_prefix_len >= 1:
            add_message(branches, same_prefix_len, node, forward=True)
            if any(same_prefix_len == len(b) for b in branches):
                return

        same_suffix_len = get_stmts_difference(branches, forward=False)
        if same_suffix_len >= 1:
            # allow wip early returns
            if same_suffix_len == 1 and isinstance(branches[0][-1], nodes.Return):
                return
            defect_node = branches[0][-1].parent

            add_message(branches, same_suffix_len, defect_node, forward=False)

    def duplicate_seq_ifs(self, node: nodes.If) -> None:
        """
        returns False iff elifs end with else
        """

        def extract_from_elif(node: nodes.If, seq_ifs: List[nodes.NodeNG]) -> bool:
            if len(node.orelse) > 0 and not node.has_elif_block():
                return False

            current = node
            while current.has_elif_block():
                elif_ = current.orelse[0]
                seq_ifs.append(elif_)
                if len(elif_.orelse) > 0 and not elif_.has_elif_block():
                    return False
                current = elif_
            return True

        def extract_from_siblings(node: nodes.If, seq_ifs: List[nodes.NodeNG]) -> None:
            sibling = node.next_sibling()
            while sibling is not None and isinstance(sibling, nodes.If):
                new: List[nodes.NodeNG] = []
                if not extract_from_elif(sibling, new):
                    return
                seq_ifs.append(sibling)
                seq_ifs.extend(new)
                sibling = sibling.next_sibling()

        def same_ifs_count(seq_ifs: List[nodes.NodeNG], start: int) -> int:
            reference = seq_ifs[start].body
            for i in range(start + 1, len(seq_ifs)):
                # do not suggest join of elif and sibling
                if seq_ifs[start].parent not in seq_ifs[i].node_ancestors():
                    return i - start

                compared = seq_ifs[i].body
                if len(reference) != len(compared):
                    return i - start
                for j in range(len(reference)):
                    if reference[j].as_string() != compared[j].as_string():
                        return i - start
            return len(seq_ifs) - start

        prev_sibling = node.previous_sibling()
        if is_parents_elif(node) or (
            isinstance(prev_sibling, nodes.If) and extract_from_elif(prev_sibling, [])
        ):
            return

        seq_ifs = [node]

        if not extract_from_elif(node, seq_ifs):
            return
        extract_from_siblings(node, seq_ifs)

        if len(seq_ifs) == 1:
            return

        i = 0
        while i < len(seq_ifs) - 1:
            count = same_ifs_count(seq_ifs, i)
            if count > 1:
                first = seq_ifs[i]
                assert isinstance(seq_ifs[i + count - 1], nodes.If)
                last = seq_ifs[i + count - 1].body[-1]

                self.add_message(
                    "duplicate-seq-ifs",
                    line=first.fromlineno,
                    col_offset=first.col_offset,
                    end_lineno=last.tolineno,
                    end_col_offset=last.end_col_offset,
                    args=(count,),
                )
            i += count

    def duplicate_exprs(self, node: nodes.Module) -> None:
        visitor = DuplicateExprVisitor()
        visitor.visit(node)

        emitted = set()

        for name, exprs in sorted(
            visitor.long_expressions.items(), key=lambda pair: -len(pair[0][0])
        ):
            if len(exprs) >= 2:
                if exprs[0].parent not in emitted:
                    expr_lines = [
                        str(expr.fromlineno)
                        for expr in sorted(exprs, key=lambda e: (e.fromlineno, e.tolineno))
                    ]
                    self.add_message(
                        "duplicate-exprs", node=exprs[0], args=(name, ", ".join(expr_lines))
                    )
                emitted.update(exprs)

    def duplicate_blocks(self, node: nodes.Module) -> None:
        def update_diffs(
            stmt1: nodes.NodeNG, stmt2: nodes.NodeNG, current_diffs: Set[Tuple[str, str]]
        ) -> bool:
            if isinstance(stmt1, nodes.Compare):
                for (op1, _), (op2, _) in zip(stmt1.ops, stmt2.ops):
                    if op1 != op2:
                        current_diffs.add((op1, op2))
            elif isinstance(stmt1, nodes.BinOp) and stmt1.op != stmt2.op:
                current_diffs.add((stmt1.op, stmt2.op))
            elif isinstance(stmt1, nodes.BoolOp) and stmt1.op != stmt2.op:
                return False
            elif isinstance(stmt1, nodes.Name) and stmt1.name != stmt2.name:
                current_diffs.add((stmt1.name, stmt2.name))
            elif isinstance(stmt1, nodes.Const) and stmt1.value != stmt2.value:
                current_diffs.add((str(stmt1.value), str(stmt2.value)))
            elif isinstance(stmt1, nodes.Attribute) and stmt1.attrname != stmt2.attrname:
                current_diffs.add((stmt1.attrname, stmt2.attrname))
            elif isinstance(stmt1, nodes.UnaryOp) and stmt1.op != stmt2.op:
                current_diffs.add((stmt1.op, stmt2.op))
            return True

        def duplicate_blocks(
            block1: List[nodes.NodeNG],
            block2: List[nodes.NodeNG],
            max_diff: int,
            current_diffs: Set[Tuple[str, str]],
        ) -> bool:
            if len(block1) != len(block2):
                return False
            for stmt1, stmt2 in zip(block1, block2):
                if not isinstance(stmt1, type(stmt2)):
                    return False

                children1 = list(stmt1.get_children())
                children2 = list(stmt2.get_children())

                if len(children1) != len(children2):
                    return False

                if not update_diffs(stmt1, stmt2, current_diffs) or len(current_diffs) > max_diff:
                    return False

                if not duplicate_blocks(children1, children2, max_diff, current_diffs):
                    return False
            return True

        MIN_LINES = 3
        MAX_DIFF = 3

        visitor = CollectBlocksVisitor()
        visitor.visit(node)

        blocks = [
            block
            for block in visitor.blocks
            if len(block) > 0
            and get_lines_between(block[0], block[-1], including_last=True) >= MIN_LINES
        ]

        if len(blocks) < 2:
            return

        blocks.sort(key=lambda block: (block[0].fromlineno, block[-1].tolineno))
        max_closed_line = 0

        for i in range(len(blocks)):
            for j in range(i + 1, len(blocks)):
                block1 = blocks[i]
                block2 = blocks[j]

                if (
                    not isinstance(block1[0].parent, type(block2[0].parent))
                    or block1[-1].tolineno <= max_closed_line
                ):
                    continue

                if duplicate_blocks(block1, block2, MAX_DIFF, set()):
                    self.add_message(
                        "duplicate-blocks",
                        line=block1[0].fromlineno,
                        col_offset=block1[0].col_offset,
                        end_lineno=block1[-1].tolineno,
                        end_col_offset=block1[-1].end_col_offset,
                        args=(f"{block1[0].fromlineno} and {block2[0].fromlineno}",),
                    )
                    max_closed_line = block1[-1].tolineno
                    break

    def duplicate_sequence(self, node: nodes.Module) -> None:
        def can_use_range(diffs: List[Any]) -> bool:
            if all(e is None for e in diffs):
                return True
            if not all(isinstance(e, int) for e in diffs):
                return False

            assert len(diffs) >= 2
            step = diffs[1] - diffs[0]
            return all(e1 + step == e2 for e1, e2 in zip(diffs, diffs[1:]))

        def get_single_diff_list(
            stmts1: List[nodes.NodeNG], stmts2: List[nodes.NodeNG]
        ) -> Optional[Tuple[Tuple[Any, Any], List[int]]]:
            if len(stmts1) != len(stmts2):
                return None

            result = None
            for i, (stmt1, stmt2) in enumerate(zip(stmts1, stmts2)):
                subresult = get_single_diff(stmt1, stmt2)
                if subresult is None or (result is not None and result[0] != subresult[0]):
                    return None

                diff, path = subresult
                if diff != (None, None):
                    path.append(i)
                    result = diff, path

            if result is None:
                return (None, None), []
            return result

        def get_single_diff(
            stmt1: nodes.NodeNG, stmt2: nodes.NodeNG
        ) -> Optional[Tuple[Tuple[Any, Any], List[int]]]:
            if not isinstance(stmt1, type(stmt2)):
                return None
            if isinstance(stmt1, nodes.Const):
                if not isinstance(stmt1.value, type(stmt2.value)):
                    return None
                if stmt1.value == stmt2.value:
                    return (None, None), []
                return (stmt1.value, stmt2.value), []
            if isinstance(stmt1, nodes.Name) and stmt1.name != stmt2.name:
                return None
            if isinstance(stmt1, nodes.Attribute) and stmt1.attrname != stmt2.attrname:
                return None
            if (
                isinstance(stmt1, (nodes.BinOp, nodes.BoolOp, nodes.UnaryOp))
                and stmt1.op != stmt2.op
            ):
                return None
            if isinstance(stmt1, nodes.Compare) and any(
                o1 != o2 for (o1, _), (o2, _) in zip(stmt1.ops, stmt2.ops)
            ):
                return None
            if isinstance(stmt1, nodes.Assign) and any(
                t1.as_string() != t2.as_string() for t1, t2 in zip(stmt1.targets, stmt2.targets)
            ):
                return None
            if (
                isinstance(stmt1, (nodes.AugAssign, nodes.AnnAssign))
                and stmt1.target.as_string() != stmt2.target.as_string()
            ):
                return None
            if is_block_comment(stmt1):
                return None
            if isinstance(stmt1, (nodes.Assert, nodes.Import, nodes.ImportFrom)):
                return None
            if (
                isinstance(stmt1, nodes.Call)
                and isinstance(stmt1.func, nodes.Name)
                and stmt1.func.name == "print"
            ):
                return None

            return get_single_diff_list(list(stmt1.get_children()), list(stmt2.get_children()))

        def get_seq_diffs(block: List[nodes.NodeNG], subblock_len: int, start: int) -> List[Any]:
            path = None
            diffs: List[Any] = []
            for i in range(start, len(block) - subblock_len, subblock_len):
                subblock1 = block[i : i + subblock_len]
                subblock2 = block[i + subblock_len : i + 2 * subblock_len]

                subresult = get_single_diff_list(subblock1, subblock2)
                if subresult is None:
                    return diffs

                diff, subpath = subresult
                if path is not None and len(path) > 0 and len(subpath) > 0 and path != subpath:
                    return diffs
                if (
                    path is not None
                    and len(path) == 0
                    and len(subpath) > 0
                    and len(diffs) >= DUPL_SEQ_LEN
                ):
                    return diffs
                if path is None or (len(path) == 0 and len(subpath) > 0):
                    path = subpath

                if len(diffs) == 0:
                    diffs.extend(diff)
                else:
                    diffs.append(diff[1])

            return diffs

        def process_block(self: "NoDuplicateCode", block: List[nodes.NodeNG]) -> None:
            max_subblock_len = len(block) // DUPL_SEQ_LEN

            if max_subblock_len == 0:
                return

            start = 0
            while start < len(block) - 1:
                for subblock_len in range(1, max_subblock_len + 1):
                    diffs = get_seq_diffs(block, subblock_len, start)
                    if (len(diffs) >= DUPL_SEQ_LEN and can_use_range(diffs)) or len(
                        diffs
                    ) >= DUPL_SEQ_LEN_NO_RANGE:
                        first_subblock = block[start : start + subblock_len]
                        last_subblock = block[
                            start
                            + (len(diffs) - 1) * subblock_len : start
                            + len(diffs) * subblock_len
                        ]
                        self.add_message(
                            "duplicate-sequence",
                            line=first_subblock[0].fromlineno,
                            col_offset=first_subblock[0].col_offset,
                            end_lineno=last_subblock[-1].tolineno,
                            end_col_offset=last_subblock[-1].end_col_offset,
                            args=(
                                len(diffs),
                                get_lines_between(
                                    first_subblock[0], first_subblock[-1], including_last=True
                                ),
                            ),
                        )
                        start += len(diffs) * subblock_len
                        break
                else:
                    start += 1

        DUPL_SEQ_LEN = 4
        DUPL_SEQ_LEN_NO_RANGE = 5

        visitor = CollectBlocksVisitor()
        visitor.visit(node)

        blocks = [block for block in visitor.blocks if len(block) > 0]
        blocks.sort(key=lambda block: (block[0].fromlineno, block[-1].tolineno))

        for block in blocks:
            if len(block) >= 2:
                process_block(self, block)

    @only_required_for_messages("duplicate-exprs", "duplicate-blocks", "duplicate-sequence")
    def visit_module(self, node: nodes.Module) -> None:
        self.duplicate_exprs(node)
        self.duplicate_blocks(node)
        self.duplicate_sequence(node)


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(NoDuplicateCode(linter))



import pytest
from edulint.options import Option
from edulint.linters import Linter
from edulint.config.arg import Arg
from edulint.linting.problem import Problem
from test_utils import lazy_problem, apply_and_lint, create_apply_and_lint
from typing import List


@pytest.mark.parametrize("filename,expected_output", [
    ("013165-coins.py", [
        lazy_problem().set_line(9)
        .set_text("Identical code inside all if's branches, move 1 lines before the if.")
    ]),
    ("015726-prime.py", [
        # lazy_problem().set_line(10)
        # .set_text("Identical code inside all if's branches, move 1 lines after the if.")
    ]),
    ("016119-fibsum.py", [
        lazy_problem().set_line(22)
        .set_text("Identical code inside all if's branches, move 3 lines after the if.")
    ]),
    ("023171-credit.py", [
        lazy_problem().set_line(26),
        lazy_problem().set_line(38)
    ]),
    ("024056-cards.py", [
        lazy_problem().set_line(7)
    ]),
    ("024329-workdays.py", [
        lazy_problem().set_line(56)
    ]),
    ("024535-credit.py", [
        lazy_problem().set_line(23)
        .set_text("Identical code inside all if's branches, move 3 lines after the if."),
        # lazy_problem().set_line(35),
    ]),
    ("052786-course.py", []),
    ("054050-course.py", [
        lazy_problem().set_line(31)
    ]),
    ("074168-doubly_linked.py", [
        lazy_problem().set_line(31)  # dubious
    ]),
    ("074242-tortoise.py", [
        lazy_problem().set_line(32),
        lazy_problem().set_line(32),
        lazy_problem().set_line(65),
        lazy_problem().set_line(65)
    ]),
    ("124573-lists.py", [
        # lazy_problem().set_line(51)
        # .set_text("Identical code inside all if's branches, move 1 lines after the if.")
    ]),
    ("hw21739.py", [  # TODO more precise detection
        lazy_problem().set_line(40)
        .set_text("Identical code inside all if's branches, move 9 lines after the if."),
        lazy_problem().set_line(42)
        .set_text("Identical code inside all if's branches, move 1 lines after the if."),
        # lazy_problem().set_line(67)
        # .set_text("Identical code inside all if's branches, move 1 lines after the if."),
        # lazy_problem().set_line(87)
        # .set_text("Identical code inside all if's branches, move 2 lines before the if."),
        # lazy_problem().set_line(87)
        # .set_text("Identical code inside all if's branches, move 3 lines after the if."),
        lazy_problem().set_line(109)
        .set_text("Identical code inside all if's branches, move 16 lines after the if."),
        lazy_problem().set_line(164)
        .set_text("Identical code inside all if's branches, move 2 lines before the if."),
        lazy_problem().set_line(164)
        .set_text("Identical code inside all if's branches, move 3 lines after the if.")
    ]),
    ("hw48505.py", []),
    ("uc_92_9765_21_23.py", [lazy_problem().set_line(2)]),
])
def test_identical_if_branches_part(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=identical-if-branches-part")],
        expected_output
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("810ed6038b-hw5.py", [lazy_problem().set_line(91)]),
    ("uc_88_6816_15_15.py", [lazy_problem().set_line(5)]),
])
def test_identical_if_branches(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=identical-if-branches")],
        expected_output
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("012889-geometry.py", [
        lazy_problem().set_line(28).set_end_line(33)
        .set_text("Identical code inside 3 consecutive ifs, join their conditions using 'or'.")
    ]),
    ("012889-pythagorean.py", [
        lazy_problem().set_line(26).set_end_line(29)
        .set_text("Identical code inside 2 consecutive ifs, join their conditions using 'or'.")
    ]),
    ("014341-geometry.py", [
        lazy_problem().set_line(9).set_end_line(14)
    ]),
    ("014613-next.py", [
        lazy_problem().set_line(35).set_end_line(40)
    ]),
    ("034440-cellular.py", [
        lazy_problem().set_line(56).set_end_line(61),
        lazy_problem().set_line(62).set_end_line(65)
    ]),
    ("046586-person_id.py", [
        lazy_problem().set_line(40).set_end_line(56)
        .set_text("Identical code inside 7 consecutive ifs, join their conditions using 'or'.")
    ]),
    ("052975-parse_time.py", []),
])
def test_identical_seq_ifs(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=identical-seq-ifs")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("012853-geometry.py", [  # TODO detect this
        lazy_problem().set_line(8)
        .set_text("A complex expression 'not (a + b > c and b + c > a and c + a > b)' used repeatedly "
                  "(on lines 8, 18, 27). Extract it to a local variable or create a helper function.")
    ]),
    ("014024-pythagorean.py", [
        lazy_problem().set_line(19)
        .set_text("A complex expression 'sqrt(a**2 + b**2)' used repeatedly (on lines 19, 20). Extract it to a "
                  "local variable or create a helper function."),
        lazy_problem().set_line(21)
        .set_text("A complex expression 'a + b + c > result' used repeatedly (on lines 21, 27, 33). Extract it to a "
                  "local variable or create a helper function."),
        lazy_problem().set_line(25),
        lazy_problem().set_line(31),
    ]),
    ("042643-edge_detection.py", [])
])
def test_duplicate_exprs(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [
            Arg(Option.PYLINT, "--load-plugins=edulint.linting.checkers.old_duplication_checker"),
            Arg(Option.PYLINT, "--enable=duplicate-exprs")
        ],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("013145-coins.py", [
        lazy_problem().set_line(5).set_end_line(7)
        .set_text("Duplicate blocks starting on lines 5 and 10. Extract the code to a helper function."),
        lazy_problem().set_line(10).set_end_line(12),
        lazy_problem().set_line(15).set_end_line(17),
        lazy_problem().set_line(20).set_end_line(22),
    ]),
    ("024508-cards.py", [
        lazy_problem().set_line(61).set_end_line(66),
    ])
])
def test_duplicate_blocks(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [
            Arg(Option.PYLINT, "--load-plugins=edulint.linting.checkers.old_duplication_checker"),
            Arg(Option.PYLINT, "--enable=duplicate-blocks")
        ],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("010666-coins.py", [
        lazy_problem().set_line(13).set_end_line(27)
        .set_text("Duplicate sequence of 5 repetitions of 3 lines of code. Use a loop to avoid this."),
    ]),
    ("014212-coins.py", [
        lazy_problem().set_line(8).set_end_line(19)
        .set_text("Duplicate sequence of 6 repetitions of 2 lines of code. Use a loop to avoid this."),
    ]),
    ("051796-gps.py", []),
    ("hw36211.py", []),
    ("ut156_9802_29_10.py", [
        lazy_problem().set_line(7).set_end_line(14)
        .set_text("Duplicate sequence of 4 repetitions of 2 lines of code. Use a loop to avoid this."),
        lazy_problem().set_line(15).set_end_line(24),
        lazy_problem().set_line(25).set_end_line(36),
        lazy_problem().set_line(37).set_end_line(48),
    ]),
])
def test_duplicate_sequence(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [
            Arg(Option.PYLINT, "--load-plugins=edulint.linting.checkers.old_duplication_checker"),
            Arg(Option.PYLINT, "--enable=duplicate-sequence")
        ],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("0ba2fdc810-p6_cellular.py", []),  # not good advice
    ("730a1a0d05-p6_workdays.py", []),  # not good advice
    ("7e1dd5c338-p2_tortoise.py", []),
    ("a5c583d460-hw1.py", []),
    ("c69c0646b0-p4_triangle.py", []),
    ("d8b80cabe6-p6_workdays.py", []),  # is in if
    ("fdc1570861-p6_workdays.py", []),  # is in if
    ("hw34406.py", []),  # is in if
    ("uc_4_2117_13_17.py", []),
])
def test_similar_block_to_function(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-block-to-function")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("0642a5c1d7-hw4.py", []),  # dubious
    ("117cb0510a-midterm.py", []),
    ("1687aeed39-hw4.py", []),
    ("7e1dd5c338-p2_tortoise.py", []),
    ("APS_Automated_Data_Reporting.py", []),
    ("b3b13aa3f7-p5_merge.py", [
        lazy_problem().set_line(37)
        .set_text("There are 3 repetitions of 7 similar statements, which can be simplified using a loop. Consider iterating over 'range(3)'.")
    ]),
    ("cf_1462_e_86.py", []),
    ("cf_20_b_30.py", []),
    ("cf_373_a_21.py", [
        lazy_problem().set_line(17)
        .set_text("There are 9 repetitions of 1 similar statements, which can be simplified using a loop. Consider iterating over 'range(9)'.")
    ]),
    ("uc_10_7828_23_16.py", []),
    ("uc_4_0123_22_08.py", [lazy_problem().set_line(4).set_end_line(7)]),
    ("uc_52_2125_16_10.py", []),
    ("uc_94_2813_13_57.py", []),
    ("ut_57_4473_30_10.py", []),  # does not suggest to-loop on mergeable loops
    ("ut_57_5508_21_10.py", [lazy_problem().set_line(1)]),
    ("ut_57_9336_15_20.py", [lazy_problem().set_line(1)]),
    ("ut_80_2230_13_11.py", []),
    ("ut_80_3906_24_11.py", [lazy_problem().set_source(Linter.PYLINT)]),  # report only one missing loop, others are misshapen
    ("ut_80_8916_12_20.py", [lazy_problem().set_source(Linter.PYLINT)]),
    ("ut_98_8463_20_35.py", [lazy_problem().set_source(Linter.PYLINT)]),
    ("voter_pennant_twirl_mayday.py", [lazy_problem().set_line(36).set_text("There are 6 repetitions of 1 similar statements, which can be simplified using a loop. Consider iterating over 'range(6)'.")]),
])
def test_similar_block_to_loop_range(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-block-to-loop-range")],
        expected_output
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("0642a5c1d7-hw4.py", []),  # dubious
    ("117cb0510a-midterm.py", [
        lazy_problem().set_line(6).set_end_line(17)
        .set_text("There are 3 repetitions of 4 similar statements, which can be simplified using a loop. Consider iterating over '[2, 3, 5]'.")
    ]),
    ("1687aeed39-hw4.py", []),
    ("7e1dd5c338-p2_tortoise.py", []),
    ("APS_Automated_Data_Reporting.py", []),
    ("b3b13aa3f7-p5_merge.py", []),
    ("cf_1462_e_86.py", []),
    ("cf_20_b_30.py", []),
    ("cf_373_a_21.py", []),
    ("uc_10_7828_23_16.py", []),
    ("uc_4_0123_22_08.py", []),
    ("uc_52_2125_16_10.py", []),
    ("uc_94_2813_13_57.py", [lazy_problem().set_line(6).set_end_line(11).set_text("There are 6 repetitions of 1 similar statements, which can be simplified using a loop. Consider iterating over '['a', 'e', 'i', 'o', 'u', 'y']'.")]),
    ("ut_57_4473_30_10.py", []),  # does not suggest to-loop on mergeable loops
    ("ut_57_5508_21_10.py", []),
    ("ut_57_9336_15_20.py", []),
    ("ut_80_2230_13_11.py", []),
    ("ut_80_3906_24_11.py", []),  # report only one missing loop, others are misshapen
    ("ut_80_8916_12_20.py", []),
    ("ut_98_8463_20_35.py", []),
    ("voter_pennant_twirl_mayday.py", []),
])
def test_similar_block_to_loop_collection(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-block-to-loop-collection")],
        expected_output
    )

@pytest.mark.parametrize("lines,expected_output", [
    ([  # 0
        "f(0)",
        "f(1)",
        "f(2)",
        "f(3)",
        "f(4)",
    ], [lazy_problem().set_line(1)]),
    ([  # 1
        "f(v + 0)",
        "f(v + 1)",
        "f(v + 2)",
        "f(v + 3)",
        "f(v + 4)",
    ], [lazy_problem().set_line(1)]),
    ([  # 2
        "f(v)",
        "f(v + 1)",
        "f(v + 2)",
        "f(v + 3)",
        "f(v + 4)",
    ], [lazy_problem().set_line(1)]),
    ([  # 3
        "f(v)",
        "f(v * 2)",
        "f(v * 3)",
        "f(v * 4)",
        "f(v * 5)",
    ], [lazy_problem().set_line(1)]),
    ([  # 4
        "f(0)",
        "f(v)",
        "f(v * 2)",
        "f(v * 3)",
        "f(v * 4)",
    ], [lazy_problem().set_line(1)]),
    ([  # 5
        "f(5)",
        "f(5 + v)",
        "f(5 + v * 2)",
        "f(5 + v * 3)",
        "f(5 + v * 4)",
    ], [lazy_problem().set_line(1)]),
    ([  # 6
        "f(k)",
        "f(k + v)",
        "f(k + v * 2)",
        "f(k + v * 3)",
        "f(k + v * 4)",
    ], [lazy_problem().set_line(1)]),
    ([  # 7
        "f(0)",
        "f(v * 2)",
        "f(v * 3)",
        "f(v * 4)",
        "f(v * 5)",
    ], [lazy_problem().set_line(2)]),
])
def test_similar_block_to_loop_range_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=similar-block-to-loop-range")],
        expected_output
    )

@pytest.mark.parametrize("lines,expected_output", [
    ([  # 0
        "def fun(c1, c2):",
        "    if c1:",
        "        if c2:",
        "            print('hello')",
        "            print('kind')",
        "        else:",
        "            print('cruel')",
        "            print('world')",
        "    else:",
        "        if c2:",
        "            print('cruel')",
        "            print('world')",
        "        else:",
        "            print('hello')",
        "            print('kind')",

    ], [
        lazy_problem().set_line(2)
        .set_text("The branches of the 'if' statement are similar. Use condition 'c1 and c2 or not c1 and not c2' to avoid the duplication.")
    ]),
])
def test_similar_if_to_untwisted_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=similar-if-to-untwisted")],
        expected_output
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("cf_1316_b_41.py", [
        lazy_problem().set_line(12)
        .set_text("The branches of the 'if' statement are similar. Use condition 'n % 2 == 0 and i % 2 == 0 or not n % 2 == 0 and not i % 2 == 0' to avoid the duplication.")
    ]),
])
def test_similar_if_to_untwisted(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-if-to-untwisted")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("ut_80_2861_19_63.py", [
        lazy_problem().set_line(3).set_end_line(10)
        .set_text("The body of the loop contains duplicate blocks, starting on lines 3, 7. Avoid the duplication by repeating the outer loop 8 times.")
    ]),
    ("ut_57_4473_30_10.py", [lazy_problem().set_line(2)]),
])
def test_similar_block_to_loop_merge(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-block-to-loop-merge")],
        expected_output
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("051746-gps.py", [lazy_problem().set_line(71)]),
    ("c93682fee8-p4_digits.py", []),
    ("cf_1440_c1_2.py", []), # TODO lazy_problem().set_line(50)
    ("ut_80_0402_13_19.py", [
        lazy_problem().set_line(8).set_end_line(10),
        # remaining are also true, but currently overriden by loop suggestion
        # lazy_problem().set_line(12),
        # lazy_problem().set_line(16),
        # lazy_problem().set_line(20),
        # lazy_problem().set_line(24),
        # lazy_problem().set_line(28),
    ]),
    ("ut_92_5508_03_10.py", [lazy_problem().set_line(7)]),
])
def test_similar_block_to_call(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-block-to-call")],
        expected_output
    )


# TODO later, currently too ambitions but probably too rare
# @pytest.mark.parametrize("lines,expected_output", [
#     ([  # 0
#         "def fun(x):",
#         "    a(x)",
#         "    b(x)",
#         "x = 5",
#         "a(x)",
#         "b(x)"
#     ], [lazy_problem().set_line(5)]),
#     ([  # 1
#         "def fun(x):",
#         "    a(x)",
#         "    b(x)",
#         "a(5)",
#         "b(5)"
#     ], [lazy_problem().set_line(4)]),
#     ([  # 2
#         "def fun(x):",
#         "    a(x)",
#         "    b(x)",
#         "lst = []",
#         "a(lst[0])",
#         "b(lst[0])"
#     ], [lazy_problem().set_line(5)]),
#     ([  # 3
#         "def fun(x):",
#         "    a(x)",
#         "    b(x)",
#         "lst = []",
#         "a(lst[0])",
#         "b(lst[1])"
#     ], []),
#     ([  # 4
#         "def fun(x):",
#         "    y = a(x)",
#         "    z = b(y)",
#         "    return z"
#         "p = a(5)",
#         "q = b(p)",
#         "print(q)"
#     ], [lazy_problem().set_line(5)]),
#     ([  # 5
#         "def fun(x):",
#         "    y = a(x)",
#         "    z = b(y)",
#         "    return int(z)"
#         "p = a(5)",
#         "q = b(p)",
#         "print(int(q))"
#     ], [lazy_problem().set_line(5)]),
#     ([  # 6
#         "def fun(x):",
#         "    y = a(x)",
#         "    z = b(y)",
#         "    return int(z) + 1"
#         "p = a(5)",
#         "q = b(p)",
#         "print(int(q))"
#     ], []),
#     ([  # 7
#         "def fun(x):",
#         "    y = a(x)",
#         "    z = b(y)",
#         "    return z",
#         "x = 5",
#         "y = a(x)",
#         "z = b(y)",
#         "print(z)"
#     ], [lazy_problem().set_line(6)]),
# ])
# def test_similar_block_to_call_custom(lines: List[str], expected_output: List[Problem]) -> None:
#     create_apply_and_lint(
#         lines,
#         [Arg(Option.PYLINT, "--enable=similar-block-to-call")],
#         expected_output
#     )

@pytest.mark.parametrize("filename,expected_output", [
    ("0bf69cc1a5-p4_geometry.py", []),
    ("163aadb1dd-p4_geometry.py", []),  # dubious
    ("53647f4477-hw3.py", [lazy_problem().set_line(91)]),
    ("5ce8692f42-p5_fibsum.py", []),
    ("769200244d-p6_workdays.py", [lazy_problem().set_line(66)]),
    ("7e1dd5c338-p1_digit_sum.py", []), # [lazy_problem().set_line(23)]), dubious
    ("7e1dd5c338-p5_credit.py", []),
    ("9668dff756-p6_workdays.py", []),
    ("ccf4a9f103-p6_workdays.py", []),
    ("cf_1113_c_27.py", [lazy_problem().set_line(16)]),
    ("cf_1271_b_32_tweaked.py", [lazy_problem().set_line(17)]),
    ("custom_if_calls_to_variables.py", []),
    ("fd637a2984-p6_workdays.py", []),
    ("fdc1570861-p6_workdays.py", [lazy_problem().set_line(47)]),  # multiple if branches differing in one value
    ("tarot_card_reader.py", []),
    ("uc_13_1073_17_21.py", [lazy_problem().set_line(2)]),
    ("uc_73_0198_15_17.py", [lazy_problem().set_line(6)]),
    ("uc_73_2551_11_17.py", [lazy_problem().set_line(3)]),
    ("uc_73_3819_50_56.py", []),
    ("uc_73_3819-20_56.py", []),
    ("uc_73_3897_10_43.py", [lazy_problem().set_line(3)]),
    ("uc_73_5468_12_52.py", [lazy_problem().set_line(3)]),
    ("uc_73_7863_14_44.py", [lazy_problem().set_line(3)]),
    ("uc_73_8593_19_21.py", []),
    ("uc_74_7627_27_19.py", [lazy_problem().set_line(6), lazy_problem().set_line(14)]),
])
def test_similar_if_to_variables(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-if-to-extracted")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("0bf69cc1a5-p4_geometry.py", []),
    ("163aadb1dd-p4_geometry.py", []),
    ("3867ee9889-split.py", []),
    ("5ce8692f42-p5_fibsum.py", []),
    ("769200244d-p6_workdays.py", [
        lazy_problem().set_line(66)
        .set_text("The branches of the 'if' statement are similar. Use if expression to decide between the different values (i.e., \"29 if gap_year(year) else 28\"), but only if the if's body has no effect on the condition's value.")
    ]),
    ("7e1dd5c338-p1_digit_sum.py", []),
    ("7e1dd5c338-p5_credit.py", []),
    ("9668dff756-p6_workdays.py", []),
    ("ccf4a9f103-p6_workdays.py", []),
    ("custom_if_calls_to_variables.py", []),
    ("fd637a2984-p6_workdays.py", []),
    ("fdc1570861-p6_workdays.py", []),
    ("tarot_card_reader.py", []),
    ("uc_40_4468_31_23.py", []),
    ("uc_56_4519_12_10.py", []),
    ("uc_73_0198_15_17.py", []),
    ("uc_73_2551_11_17.py", [
        lazy_problem().set_line(3)
        .set_text("The branches of the 'if' statement are similar. Use if expression to decide between the different values (i.e., \"'#' if i // m % 2 == 0 else '.'\"), but only if the if's body has no effect on the condition's value.")
    ]),
    ("uc_73_3819_50_56.py", [lazy_problem().set_line(6), lazy_problem().set_line(16)]),
    ("uc_73_3819-20_56.py", [lazy_problem().set_line(7), lazy_problem().set_line(18)]),
    ("uc_73_3897_10_43.py", [
        lazy_problem().set_line(3).set_text("The branches of the 'if' statement are similar. Use if expression to decide between the different values (i.e., \"0 if i % 2 == 0 else 1\"), but only if the if's body has no effect on the condition's value.")
    ]),
    ("uc_73_5468_12_52.py", []),
    ("uc_73_7863_14_44.py", []),
    ("uc_73_8593_19_21.py", []),
    ("uc_82_1702_19_11.py", []),
    ("uc_87_5303_19_15.py", []),
])
def test_similar_if_to_expr(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-if-to-expr")],
        expected_output
    )


@pytest.mark.parametrize("lines,expected_output", [
    ([  # 0
        "def fun(c1, c2):",
        "    if c1:",
        "        if c2:",
        "            print('hello')",
        "            print('kind')",
        "        else:",
        "            print('cruel')",
        "            print('world')",
        "    else:",
        "        if c2:",
        "            print('hello')",
        "            print('foo')",
        "        else:",
        "            print('cruel')",
        "            print('world')",

    ], [
        lazy_problem().set_line(2)
        .set_text("The branches of the 'if' statement are similar. Move the 'if' into the branches, so that it only contains lines 5, 12, but only if the if's body has no effect on the condition's value.")
    ]),
])
def test_similar_if_into_block_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=similar-if-into-block")],
        expected_output
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("0bf69cc1a5-p4_geometry.py", []),
    ("163aadb1dd-p4_geometry.py", []),
    ("2a6a841f84-task_2.py", [
        lazy_problem().set_line(36)
        .set_text("The branches of the 'if' statement are similar. Move the 'if' into the branches, so that it only contains lines 40, 47, but only if the if's body has no effect on the condition's value.")
    ]),
    ("30c8bd0ad7-p6_cellular.py", []),
    ("3867ee9889-split.py", []),
    ("5ce8692f42-p5_fibsum.py", []),
    ("769200244d-p6_workdays.py", [
        lazy_problem().set_line(66)
        .set_text("The branches of the 'if' statement are similar. Move the 'if' into the branches, so that it only contains lines 75-76, 88-89, but only if the if's body has no effect on the condition's value.")
    ]),
    ("7e1dd5c338-p1_digit_sum.py", []),
    ("7e1dd5c338-p5_credit.py", []),
    ("9668dff756-p6_workdays.py", []),
    ("ccf4a9f103-p6_workdays.py", []),
    ("cf_1503_a_14.py", []),
    ("cf_479_c_0.py", [lazy_problem().set_line(12)]),
    ("cf_621_b_47.py", []),
    ("cf_911_e_11.py", []),
    ("custom_if_calls_to_variables.py", []),
    ("fd637a2984-p6_workdays.py", []),
    ("fdc1570861-p6_workdays.py", []),
    ("ksi_12_136_aatb.py", [
        lazy_problem().set_line(4)
        .set_text("The branches of the 'if' statement are similar. Move the 'if' into the branches, so that it only contains lines 6, 11, but only if the if's body has no effect on the condition's value.")
    ]),
    ("tarot_card_reader.py", []),
    ("uc_28_6710_05_14.py", []),
    ("uc_4380_12_21.py", []),
    ("uc_73_0198_15_17.py", [lazy_problem().set_line(6)]), # dubious
    ("uc_73_2551_11_17.py", [lazy_problem().set_line(3)]),
    ("uc_73_3819_50_56.py", [
        lazy_problem().set_line(3)
        .set_text("The branches of the 'if' statement are similar. Move the 'if' into the branches, so that it only contains lines 6-11, 16-21, but only if the if's body has no effect on the condition's value.")
    ]), # multi-step
    ("uc_73_3819-20_56.py", [lazy_problem().set_line(3)]), #, lazy_problem().set_line(7)]),
    ("uc_73_3897_10_43.py", [lazy_problem().set_line(3)]), # multi-step
    ("uc_73_5468_12_52.py", []),
    ("uc_73_7863_14_44.py", [lazy_problem().set_line(3)]), # multi-step
    ("uc_73_8593_19_21.py", []),
])
def test_similar_if_into_block(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-if-into-block")],
        expected_output
    )

@pytest.mark.parametrize("filename,expected_output",[
    ("d2e87302a6-b_family_tree.py", [lazy_problem().set_line(82)]),
    ("014823-p4_geometry.py", []),
    ("m2630.py", []),
])
def test_similar_if_to_use(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=similar-if-to-use")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("0379f32d24-p6_workdays.py", [lazy_problem().set_symbol("similar-block-to-loop-collection").set_line(75)]),
    ("406561e0ae-workdays.py", []),  # contains duplication, but no good advice
    ("53647f4477-hw3.py", [
        lazy_problem().set_symbol("similar-if-to-extracted").set_line(91),
        lazy_problem().set_symbol("similar-if-to-expr").set_line(105)
    ]),
    ("63257195a0-p6_workdays.py", []),  # contains duplication, but no good advice
    ("cf_1086_a_14.py", []),
    ("cf_1113_c_27.py", [lazy_problem().set_symbol("similar-if-to-extracted").set_line(16)]),
    ("cf_1399_e_0.py", [lazy_problem().set_symbol("similar-if-into-block").set_line(44)]),
    ("cf_1551_d_11.py", [
        lazy_problem().set_symbol("similar-if-to-untwisted").set_line(17),
        lazy_problem().set_symbol("similar-if-to-untwisted").set_line(34),
        lazy_problem().set_symbol("similar-if-to-untwisted").set_line(51),
        lazy_problem().set_symbol("similar-if-to-expr").set_line(70),
        lazy_problem().set_symbol("similar-if-to-untwisted").set_line(82),
        lazy_problem().set_symbol("similar-if-to-untwisted").set_line(99),
        lazy_problem().set_symbol("similar-if-to-untwisted").set_line(116),
    ]),
    # contains more duplication, but no good advice
    ("cf_994_c_10.py", [lazy_problem().set_symbol("identical-seq-ifs").set_line(26)]),
    ("fdc1570861-p6_workdays.py", [lazy_problem().set_symbol("similar-if-to-extracted").set_line(47)]),
    # TODO suggest moving print into the for below
    ("uc_11_8337_17_22.py", [lazy_problem().set_symbol("similar-if-to-extracted").set_line(3)]),
    ("uc_13_1073_17_21.py", [lazy_problem().set_symbol("similar-if-to-extracted").set_line(2)]),
    ("uc_28_6562_15_18.py", [lazy_problem().set_symbol("similar-if-to-extracted").set_line(4)]),
    ("uc_28_7857_02_17.py", [lazy_problem().set_symbol("similar-if-to-extracted").set_line(3)]),
    ("uc_40_4468_31_23.py", []),
    # TODO suggest breaking up the string
    ("uc_56_4519_12_10.py", []),
    ("uc_73_6644_16_22.py", [lazy_problem().set_symbol("similar-if-into-block").set_line(14)]),
    ("uc_74_7627_27_19.py", [lazy_problem().set_line(6), lazy_problem().set_line(14)]),
    # TODO suggest breaking up the string
    ("uc_82_1702_19_11.py", []),
    ("uc_87_5303_19_15.py", []),  # dubious
])
def test_interactions(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=no-duplicate-code")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("cf_1166_c_4.py", []),
    ("ksi_17_513_aacd.py", []),
    # TODO report identical functions
    ("pronto_jawless_seismic_hefty.py", []),
])
def test_cfg(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [],
        expected_output
    )



from astroid import nodes  # type: ignore
from typing import TYPE_CHECKING, Optional, List, Tuple, Union, Any

from pylint.checkers import BaseChecker
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter

from edulint.linting.checkers.utils import (
    get_range_params,
    get_const_value,
    infer_to_value,
    is_parents_elif,
    get_statements_count,
    is_negation,
)


class Local(BaseChecker):
    name = "local-defects"
    msgs = {
        "R6600": (
            "Should never be emitted",
            "noop",
            "Never emitted, just to use if all other checks are disabled",
        ),
        "R6601": (
            "Use %s.append(%s) instead of %s.",
            "use-append",
            "Emitted when code extends list by a single argument instead of appending it.",
        ),
        "R6602": (
            "Use integral division //.",
            "use-integral-division",
            "Emitted when the code uses float division and converts the result to int.",
        ),
        "R6603": (
            "Use isdecimal to test if string contains a number.",
            "use-isdecimal",
            "Emitted when the code uses isdigit or isnumeric.",
        ),
        "R6604": (
            "Do not use %s loop with else.",
            "no-loop-else",
            "Emitted when the code contains loop with else block.",
        ),
        "R6605": (
            "Use elif.",
            "use-elif",
            "Emitted when the code contains else: if construction instead of elif. Might return a false positive if "
            "the code mixes tabs and spaces.",
        ),
        "R6606": (
            "The for loop makes %s.",
            "at-most-one-iteration-for-loop",
            "Emitted when a for loop would always perform at most one iteration.",
        ),
        "R6607": (
            "Use %s instead of repeated %s in %s.",
            "no-repeated-op",
            "Emitted when the code contains repeated adition/multiplication instead of multiplication/exponentiation.",
        ),
        "R6608": (
            "Redundant arithmetic: %s",
            "redundant-arithmetic",
            "Emitted when there is redundant arithmetic (e.g. +0, *1) in an expression.",
        ),
        "R6609": (
            "Use augmented assignment: '%s %s= %s'",
            "use-augmented-assign",
            "Emitted when an assignment can be simplified by using its augmented version.",
        ),
        "R6610": (
            "Do not multiply list with mutable content.",
            "do-not-multiply-mutable",
            "Emitted when a list with mutable contents is being multiplied.",
        ),
        "R6611": (
            "Use else instead of elif.",
            "redundant-elif",
            "Emitted when the condition in elif is negation of the condition in the if.",
        ),
        "R6612": (
            "Unreachable else.",
            "unreachable-else",
            "Emitted when the else branch is unreachable due to totally exhaustive conditions before.",
        ),
        "R6613": (
            "Use '%s' directly rather than as '%s'.",
            "no-is-bool",
            "Emitted when the is operator is used with a bool value.",
        ),
        "R6614": (
            'Use "%s" instead of using the magical constant %i.',
            "use-ord-letter",
            "Emitted when the code uses a magical constant instead of a string literal.",
        ),
        "R6615": (
            "Remove the call to 'ord' and compare to the string directly \"%s\" instead of using "
            "the magical constant %i. Careful, this may require changing the comparison operator.",
            "use-literal-letter",
            "Emitted when the code uses a magical constant instead of a string literal in a comparison.",
        ),
        "R6616": (
            "Use early return.",
            "use-early-return",
            "Emitted when a long block of code is followed by an else that just returns, breaks or continues.",
        ),
    }

    def _check_extend(self, node: nodes.Call) -> None:
        if (
            isinstance(node.func, nodes.Attribute)
            and node.func.attrname == "extend"
            and len(node.args) == 1
            and isinstance(node.args[0], nodes.List)
            and len(node.args[0].elts) == 1
        ):
            self.add_message(
                "use-append",
                node=node,
                args=(
                    node.func.expr.as_string(),
                    node.args[0].elts[0].as_string(),
                    node.as_string(),
                ),
            )

    def _check_augassign_extend(self, node: nodes.AugAssign) -> None:
        if node.op == "+=" and isinstance(node.value, nodes.List) and len(node.value.elts) == 1:
            self.add_message(
                "use-append",
                node=node,
                args=(node.target.as_string(), node.value.elts[0].as_string(), node.as_string()),
            )

    def _check_isdecimal(self, node: nodes.Call) -> None:
        if isinstance(node.func, nodes.Attribute) and node.func.attrname in (
            "isdigit",
            "isnumeric",
        ):
            self.add_message("use-isdecimal", node=node)

    def _check_div(self, node: nodes.Call) -> None:
        if (
            isinstance(node.func, nodes.Name)
            and node.func.name == "int"
            and len(node.args) == 1
            and isinstance(node.args[0], nodes.BinOp)
            and node.args[0].op == "/"
        ):
            self.add_message("use-integral-division", node=node)

    def _check_loop_else(self, nodes: List[nodes.NodeNG], parent_name: str) -> None:
        if nodes:
            self.add_message("no-loop-else", node=nodes[0].parent, args=(parent_name))

    def _check_else_if(self, node: nodes.If) -> None:
        if node.has_elif_block():
            first_body = node.body[0]
            first_orelse = node.orelse[0]
            if first_body.col_offset == first_orelse.col_offset:
                self.add_message("use-elif", node=node.orelse[0])

    def _check_iteration_count(self, node: nodes.For) -> None:
        def get_const(node: nodes.NodeNG) -> Any:
            return node.value if isinstance(node, nodes.Const) else None

        range_params = get_range_params(node.iter)
        if range_params is None:
            return

        start, stop, step = range_params
        start, stop, step = get_const(start), get_const(stop), get_const(step)

        if start is not None and stop is not None and step is not None:
            if start >= stop:
                self.add_message(
                    "at-most-one-iteration-for-loop", node=node, args=("no iterations",)
                )
            elif start + step >= stop:
                self.add_message(
                    "at-most-one-iteration-for-loop", node=node, args=("only one iteration",)
                )

    def _check_repeated_operation_rec(
        self, node: nodes.NodeNG, op: str, name: Optional[str] = None
    ) -> Optional[Tuple[int, str]]:
        if isinstance(node, nodes.BinOp):
            if node.op != op:
                return None

            lt = self._check_repeated_operation_rec(node.left, op, name)
            if lt is None:
                return None

            count_lt, name_lt = lt
            assert name is None or name == name_lt
            rt = self._check_repeated_operation_rec(node.right, op, name_lt)
            if rt is None:
                return None

            count_rt, _ = rt
            return count_lt + count_rt, name

        if (
            name is None and type(node) in (nodes.Name, nodes.Attribute, nodes.Subscript)
        ) or name == node.as_string():
            return 1, node.as_string()
        return None

    def _check_repeated_operation(self, node: nodes.BinOp) -> None:
        if node.op in ("+", "*"):
            result = self._check_repeated_operation_rec(node, node.op)
            if result is None:
                return

            # DANGER: on some structures, + may be available but not *
            self.add_message(
                "no-repeated-op",
                node=node,
                args=(
                    "multiplication" if node.op == "+" else "exponentiation",
                    "addition" if node.op == "+" else "muliplication",
                    node.as_string(),
                ),
            )

    def _check_redundant_arithmetic(self, node: Union[nodes.BinOp, nodes.AugAssign]) -> None:
        if isinstance(node, nodes.BinOp):
            op = node.op
            left = get_const_value(node.left)
            right = get_const_value(node.right)
        elif isinstance(node, nodes.AugAssign):
            op = node.op[:-1]
            left = None
            right = get_const_value(node.value)
        else:
            assert False, "unreachable"

        if (
            (op == "+" and (left in (0, "") or right in (0, "")))
            or (op == "-" and (left == 0 or right == 0))
            or (op == "*" and (left in (0, 1) or right in (0, 1)))
            or (op == "/" and right == 1)
            or (
                op in ("/", "//", "%")
                and (
                    isinstance(node, nodes.BinOp)
                    and node.left.as_string() == node.right.as_string()
                    or isinstance(node, nodes.AugAssign)
                    and node.target.as_string() == node.value.as_string()
                )
            )
            or (op == "**" and right in (0, 1))
        ):
            self.add_message("redundant-arithmetic", node=node, args=(node.as_string(),))

    def _check_augmentable(self, node: Union[nodes.Assign, nodes.AnnAssign]) -> None:
        IMMUTABLE_OPS = ("/", "//", "%", "**", "<<", ">>")

        def add_message(target: str, param: nodes.BinOp) -> None:
            self.add_message(
                "use-augmented-assign",
                node=node,
                args=(target, node.value.op, param.as_string()),
            )

        def is_immutable(node: nodes.NodeNG) -> bool:
            if isinstance(node, (nodes.Const)) and isinstance(
                node.value, (int, float, bool, str, bytes, tuple)
            ):
                return True
            if isinstance(node, nodes.BinOp):
                return (
                    node.op in IMMUTABLE_OPS or is_immutable(node.left) or is_immutable(node.right)
                )
            if isinstance(node, nodes.Call):
                return any(
                    node.func.as_string().endswith(n)
                    for n in (
                        "int",
                        "float",
                        "bool",
                        "str",
                        "bytes",
                        "tuple",
                        "len",
                        "sum",
                        "chr",
                        "ord",
                        "trunc",
                        "round",
                        "sqrt",
                        "cos",
                        "sin",
                        "radians",
                        "degrees",
                    )
                )
            if isinstance(node, nodes.IfExp):
                return is_immutable(node.body) or is_immutable(node.orelse)
            return False

        if not isinstance(node.value, nodes.BinOp):
            return
        bin_op = node.value

        if isinstance(node, nodes.Assign):
            target = node.targets[0].as_string()
        elif isinstance(node, nodes.AnnAssign):
            target = node.target.as_string()
        else:
            assert False, "unreachable"

        left_value = infer_to_value(bin_op.left)
        right_value = infer_to_value(bin_op.right)

        if node.value.op in IMMUTABLE_OPS or is_immutable(left_value) or is_immutable(right_value):
            if target == bin_op.left.as_string():
                add_message(target, bin_op.right)
            elif bin_op.op in "+*|&" and target == bin_op.right.as_string():
                if not (
                    isinstance(left_value, nodes.Const)
                    and isinstance(left_value.value, (str, bytes, tuple))
                ) and not (
                    isinstance(right_value, nodes.Const)
                    and isinstance(right_value.value, (str, bytes, tuple))
                ):
                    add_message(target, bin_op.left)

    def _check_multiplied_list(self, node: nodes.BinOp) -> None:
        def is_mutable(elem: nodes.NodeNG) -> bool:
            return type(elem) in (nodes.List, nodes.Set, nodes.Dict) or (
                isinstance(elem, nodes.Call)
                and isinstance(elem.func, nodes.Name)
                and elem.func.name in ("list", "set", "dict")
            )

        if node.op != "*" or (
            not isinstance(node.left, nodes.List) and not isinstance(node.right, nodes.List)
        ):
            return

        assert not isinstance(node.left, nodes.List) or not isinstance(node.right, nodes.List)
        lst = node.left if isinstance(node.left, nodes.List) else node.right

        if any(is_mutable(elem) for elem in lst.elts):
            self.add_message("do-not-multiply-mutable", node=node)

    def _check_redundant_elif(self, node: nodes.If) -> None:
        if_test = node.test
        if node.has_elif_block():
            next_if = node.orelse[0]
        # TODO report another message (may be FP)
        # elif isinstance(node.next_sibling(), nodes.If) and len(node.next_sibling().orelse) == 0:
        #     next_if = node.next_sibling()
        else:
            return

        if is_negation(if_test, next_if.test, negated_rt=False):
            self.add_message("redundant-elif", node=next_if)
            if (
                node.has_elif_block()
                and next_if == node.orelse[0]
                and len(node.orelse[0].orelse) > 0
            ):
                self.add_message("unreachable-else", node=node.orelse[0].orelse[0])

    def _check_no_is(self, node: nodes.Compare) -> None:
        for i, (op, val) in enumerate(node.ops):
            if (
                op in ("is", "is not")
                and isinstance(val, nodes.Const)
                and isinstance(val.value, bool)
            ):
                prev_val = node.ops[i - 1][1] if i > 0 else node.left
                negate = (op == "is") != val.value
                self.add_message(
                    "no-is-bool",
                    node=node,
                    args=(
                        f"{'not ' if negate else ''}{prev_val.as_string()}",
                        f"{prev_val.as_string()} {op} {val.as_string()}",
                    ),
                )

    def _is_ord(self, node: nodes.NodeNG) -> bool:
        return isinstance(node, nodes.Call) and node.func.as_string() == "ord"

    def _contains_ord(self, node: nodes.NodeNG) -> bool:
        if self._is_ord(node):
            return True
        if isinstance(node, nodes.BinOp):
            return self._is_ord(node.left) or self._is_ord(node.right)
        return False

    def _is_preffered(self, value: int) -> bool:
        return chr(value) in "azAZ09"

    def _in_suggestable_range(self, value: int) -> bool:
        return ord(" ") <= value < 127  # chr(127) is weird

    def _check_use_ord_letter(self, node: nodes.BinOp) -> None:
        def add_message(param: nodes.NodeNG, value: int, suggestion: str) -> None:
            self.add_message("use-ord-letter", node=param, args=(suggestion, value))

        if node.op not in ("+", "-"):
            return

        for ord_param, const_param in ((node.left, node.right), (node.right, node.left)):
            value = get_const_value(const_param)
            if (
                value is None
                or not isinstance(value, int)
                or not self._in_suggestable_range(value)
                or not self._contains_ord(ord_param)
            ):
                continue

            if self._is_preffered(value + 1):
                suggestion = f"ord('{chr(value + 1)}') - 1"
            elif self._is_preffered(value - 1):
                suggestion = f"ord('{chr(value - 1)}') + 1"
            elif not isinstance(ord_param, nodes.BinOp) or self._is_preffered(value):
                suggestion = f"ord('{chr(value)}')"
            else:
                continue

            if (node.op == "-" and const_param == node.right and suggestion.endswith("+ 1")) or (
                node.op == "+" and const_param == node.left and suggestion.endswith("- 1")
            ):
                add_message(const_param, value, f"({suggestion})")
            else:
                add_message(const_param, value, suggestion)

    def _check_magical_constant_in_ord_compare(self, node: nodes.Compare) -> None:
        def add_message(param: nodes.NodeNG, value: int, suggestion: str) -> None:
            self.add_message("use-literal-letter", node=param, args=(suggestion, value))

        def change(op: str) -> str:
            if op == "<":
                return "<="
            if op == "<=":
                return "<"
            if op == ">":
                return ">="
            if op == ">=":
                return ">"
            assert False, "unreachable"

        all_ops = [(None, node.left)] + node.ops
        contains_ord = [self._contains_ord(param) for _, param in all_ops]

        if not any(contains_ord):
            return

        for i, (op, param) in enumerate([(None, node.left), node.ops[-1]]):
            value = get_const_value(param)
            if value is None or not self._in_suggestable_range(value):
                continue

            if i == 0:
                op, other = node.ops[0]
                if self._is_preffered(value + 1) and op in ("<", ">="):
                    add_message(param, value, f"'{chr(value + 1)}' {change(op)}")
                elif self._is_preffered(value - 1) and op in (">", "<="):
                    add_message(param, value, f"'{chr(value - 1)}' {change(op)}")
                elif (
                    not contains_ord[1]
                    or not isinstance(other, nodes.BinOp)
                    or self._is_preffered(value)
                ):
                    add_message(param, value, f"'{chr(value)}' {op}")
            else:
                if self._is_preffered(value + 1) and op in (">", "<="):
                    add_message(param, value, f"{change(op)} '{chr(value + 1)}'")
                elif self._is_preffered(value - 1) and op in ("<", ">="):
                    add_message(param, value, f"{change(op)} '{chr(value - 1)}'")
                elif (
                    not contains_ord[-2]
                    or not isinstance(all_ops[-2][1], nodes.BinOp)
                    or self._is_preffered(value)
                ):
                    add_message(param, value, f"{op} '{chr(value)}'")

    def _check_use_early_return(self, node: nodes.If):
        def ends_block(node: nodes.NodeNG) -> bool:
            if isinstance(node, nodes.If):
                return (
                    ends_block(node.body[-1])
                    and len(node.orelse) > 0
                    and ends_block(node.orelse[-1])
                )
            return isinstance(node, (nodes.Return, nodes.Break, nodes.Continue))

        def large_terminated_if_elif_block(node: nodes.NodeNG) -> bool:
            return (
                isinstance(node, nodes.If)
                and ends_block(node.body[-1])
                and get_statements_count(node.body, include_defs=True, include_name_main=True) > 2
                and (
                    len(node.orelse) == 0
                    or (len(node.orelse) == 1 and large_terminated_if_elif_block(node.orelse[0]))
                )
            )

        prev_sibling = node.previous_sibling()
        if is_parents_elif(node) or large_terminated_if_elif_block(prev_sibling):
            return

        if len(node.orelse) > 0:
            last_block = node.orelse
        elif ends_block(node.body[-1]):
            last_block = [node.next_sibling()]
        else:
            return

        if (
            ends_block(last_block[-1])
            and get_statements_count(last_block, include_defs=True, include_name_main=True) <= 2
            and get_statements_count(node.body, include_defs=True, include_name_main=True) >= 4
        ):
            self.add_message("use-early-return", node=node)

    @only_required_for_messages("use-append", "use-isdecimal", "use-integral-division")
    def visit_call(self, node: nodes.Call) -> None:
        self._check_extend(node)
        self._check_isdecimal(node)
        self._check_div(node)

    @only_required_for_messages("use-append", "redundant-arithmetic")
    def visit_augassign(self, node: nodes.AugAssign) -> None:
        self._check_augassign_extend(node)
        self._check_redundant_arithmetic(node)

    @only_required_for_messages("no-loop-else")
    def visit_while(self, node: nodes.While) -> None:
        self._check_loop_else(node.orelse, "while")

    @only_required_for_messages("no-loop-else", "at-most-one-iteration-for-loop")
    def visit_for(self, node: nodes.For) -> None:
        self._check_loop_else(node.orelse, "for")
        self._check_iteration_count(node)

    @only_required_for_messages("use-elif", "redundant-elif", "use-early-return")
    def visit_if(self, node: nodes.If) -> None:
        self._check_else_if(node)
        self._check_redundant_elif(node)
        self._check_use_early_return(node)

    @only_required_for_messages(
        "no-repeated-op", "redundant-arithmetic", "do-not-multiply-mutable", "use-ord-letter"
    )
    def visit_binop(self, node: nodes.BinOp) -> None:
        self._check_repeated_operation(node)
        self._check_redundant_arithmetic(node)
        self._check_multiplied_list(node)
        self._check_use_ord_letter(node)

    @only_required_for_messages("use-augmented-assign")
    def visit_assign(self, node: nodes.Assign) -> None:
        self._check_augmentable(node)

    @only_required_for_messages("use-augmented-assign")
    def visit_annassign(self, node: nodes.AnnAssign) -> None:
        self._check_augmentable(node)

    @only_required_for_messages("no-is-bool", "magical-constant-in-ord-compare")
    def visit_compare(self, node: nodes.Compare) -> None:
        self._check_no_is(node)
        self._check_magical_constant_in_ord_compare(node)


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(Local(linter))



from typing import List, Callable, Tuple, Dict, Set, Any
from edulint.linting.problem import ProblemJson, Problem
from edulint.linting.nonparsing_checkers import report_infile_config
from edulint.linting.process_handler import ProcessHandler
from edulint.linting.overrides import get_overriders
from edulint.linting.tweakers import get_tweakers, Tweakers
from edulint.config.config import ImmutableConfig
from edulint.config.language_translations import LangTranslations
from edulint.options import Option, ImmutableT
from edulint.linters import Linter
from functools import partial
import sys
import json
import os
from pathlib import Path
from loguru import logger


class EduLintLinterFailedException(Exception):
    pass


def get_proper_path(path: str) -> str:
    return os.path.abspath(path) if os.path.isabs(path) else os.path.relpath(path)


def flake8_to_problem(enablers: Dict[str, str], raw: ProblemJson) -> Problem:
    assert isinstance(raw["filename"], str), f'got {type(raw["filename"])} for filename'
    assert isinstance(raw["line_number"], int), f'got {type(raw["line_number"])} for line_number'
    assert isinstance(
        raw["column_number"], int
    ), f'got {type(raw["column_number"])} for column_number'
    assert isinstance(raw["code"], str), f'got {type(raw["code"])} for code'
    assert isinstance(raw["text"], str), f'got {type(raw["text"])} for text'

    return Problem(
        Linter.FLAKE8,
        enablers.get(raw["code"]),
        get_proper_path(raw["filename"]),
        raw["line_number"],
        raw["column_number"],
        raw["code"],
        raw["text"],
    )


def pylint_to_problem(
    files_or_dirs: List[str], enablers: Dict[str, str], raw: ProblemJson
) -> Problem:
    assert isinstance(raw["path"], str), f'got {type(raw["path"])} for path'
    assert isinstance(raw["line"], int), f'got {type(raw["line"])} for line'
    assert isinstance(raw["column"], int), f'got {type(raw["column"])} for column'
    assert isinstance(raw["messageId"], str), f'got {type(raw["messageId"])} for messageId'
    assert isinstance(raw["message"], str), f'got {type(raw["message"])} for message'
    assert (
        isinstance(raw["endLine"], int) or raw["endLine"] is None
    ), f'got {type(raw["endLine"])} for endLine'
    assert (
        isinstance(raw["endColumn"], int) or raw["endColumn"] is None
    ), f'got {type(raw["endColumn"])} for endColumn'
    assert isinstance(raw["symbol"], str), f'get {type(raw["symbol"])} for symbol'

    def get_used_filename(path: str) -> str:
        path = Path(path)
        abs_path = path.resolve()
        for fd in [Path(path) for path in files_or_dirs]:
            abs_fd = fd.resolve()
            if abs_fd == abs_path or abs_fd in abs_path.parents:
                if fd.is_absolute():
                    return abs_path
                else:
                    cwd_parts = Path.cwd().parts
                    abs_path_parts = abs_path.parts

                    for i in range(min(len(cwd_parts), len(abs_path_parts))):
                        if cwd_parts[i] != abs_path_parts[i]:
                            break

                    return Path(*[".."] * (len(cwd_parts) - i), *abs_path_parts[i:])
        assert False, "unreachable"

    code_enabler = enablers.get(raw["messageId"])
    symbol_enabler = enablers.get(raw["symbol"])

    return Problem(
        Linter.PYLINT,
        code_enabler if code_enabler is not None else symbol_enabler,
        get_proper_path(get_used_filename(raw["path"])),
        raw["line"],
        raw["column"],
        raw["messageId"],
        raw["message"],
        raw["endLine"],
        raw["endColumn"],
        raw["symbol"],
    )


def lint_in_subprocess(
    linter: Linter, files_or_dirs: List[str], linter_args: List[str], config_arg: ImmutableT
) -> List[Problem]:
    command = [sys.executable, "-m", str(linter)] + linter_args + list(config_arg) + files_or_dirs  # type: ignore
    return_code, outs, errs = ProcessHandler.run(command, timeout=1000)

    if ProcessHandler.is_status_code_by_timeout(return_code):
        logger.critical("{linter} was likely killed by timeout", linter=linter)
        raise TimeoutError(f"Timeout from {linter}")

    return return_code, outs, errs


def process_results(
    linter: Linter,
    return_code: int,
    outs: str,
    errs: str,
    result_getter: Callable[[Any], Any],
    out_to_problem: Callable[[Dict[str, str], ProblemJson], Problem],
    enablers: Dict[str, str],
) -> List[Problem]:
    errs = errs.strip()
    if errs:
        logger.error(errs)

    if (linter == Linter.FLAKE8 and return_code not in (0, 1)) or (
        linter == Linter.PYLINT and return_code == 32
    ):
        logger.critical(
            "{linter} exited with {return_code}", linter=linter, return_code=return_code
        )
        raise EduLintLinterFailedException()

    if not outs:
        return []

    try:
        parsed = json.loads(outs)
    except json.decoder.JSONDecodeError as e:
        logger.critical("could not parse results:\n{e}", e=e)
        logger.debug(outs)
        raise e

    return list(map(partial(out_to_problem, enablers), result_getter(parsed)))


def lint_edulint(files_or_dirs: List[str], config: ImmutableConfig) -> List[Problem]:
    ignored_infile = set(config[Option.IGNORE_INFILE_CONFIG_FOR])
    if len(ignored_infile) > 0:
        return report_infile_config(files_or_dirs, ignored_infile, config.enablers)
    return []


def lint_flake8(files_or_dirs: List[str], config: ImmutableConfig) -> List[Problem]:
    flake8_args = ["--format=json"]
    return_code, outs, errs = lint_in_subprocess(
        Linter.FLAKE8, files_or_dirs, flake8_args, config[Option.FLAKE8]
    )

    return process_results(
        Linter.FLAKE8,
        return_code,
        outs,
        errs,
        lambda r: [problem for problems in r.values() for problem in problems],
        flake8_to_problem,
        config.enablers,
    )


def lint_pylint(files_or_dirs: List[str], config: ImmutableConfig) -> List[Problem]:
    pylint_args = ["--recursive=y"] + list(config[Option.PYLINT]) + files_or_dirs

    from io import StringIO
    from pylint.lint import Run
    from pylint.reporters.json_reporter import JSON2Reporter

    output = StringIO()
    reporter = JSON2Reporter(output)

    new_stderr = StringIO()
    old_stderr = sys.stderr
    sys.stderr = new_stderr

    try:
        Run(pylint_args, reporter=reporter, exit=False)
        return_code = 0
    except SystemExit as e:
        return_code = e.code

    sys.stderr = old_stderr

    return process_results(
        Linter.PYLINT,
        return_code,
        output.getvalue(),
        new_stderr.getvalue(),
        lambda r: r["messages"],
        partial(pylint_to_problem, files_or_dirs),
        config.enablers,
    )


def apply_overrides(problems: List[Problem], overriders: Dict[str, Set[str]]) -> List[Problem]:
    codes_on_lines: Dict[int, Set[str]] = {
        line: set() for line in set([problem.line for problem in problems])
    }

    for problem in problems:
        codes_on_lines[problem.line].add(problem.code)

    result = []
    for problem in problems:
        o = overriders.get(problem.code, set())
        if not (o & codes_on_lines[problem.line]):
            result.append(problem)

    return result


def apply_tweaks(
    problems: List[Problem], tweakers: Tweakers, config: ImmutableConfig
) -> List[Problem]:
    result = []
    for problem in problems:
        tweaker = tweakers.get((problem.source, problem.code))
        if tweaker:
            if tweaker.should_keep(
                problem, [arg for arg in config if arg.option in tweaker.used_options]
            ):
                problem.text = tweaker.get_reword(problem)
                result.append(problem)
        else:
            result.append(problem)
    return result


def lint_one(filename: str, config: Tuple[ImmutableConfig, LangTranslations]) -> List[Problem]:
    option_config, lang_translations = config
    return lint_many([([filename], option_config, lang_translations)])


def sort(files_or_dirs: List[str], problems: List[Problem]) -> List[Problem]:
    files_or_dirs = [Path(path) for path in files_or_dirs]

    def get_file_or_dir_index(path: str):
        path = Path(path)
        for i, fd in enumerate(files_or_dirs):
            if fd == path or fd in path.parents:
                return i
        assert False, "unreachable"

    problems.sort(
        key=lambda problem: (
            get_file_or_dir_index(problem.path),
            problem.path,
            problem.line,
            problem.column,
        )
    )
    return problems


def lint(files_or_dirs: List[str], config: ImmutableConfig) -> List[Problem]:
    logger.info("linting files: {files_or_dirs}", files_or_dirs=files_or_dirs)
    logger.info("using config: {config}", config=config)

    edulint_result = lint_edulint(files_or_dirs, config)
    flake8_result = [] if config[Option.NO_FLAKE8] else lint_flake8(files_or_dirs, config)
    pylint_result = lint_pylint(files_or_dirs, config)

    result = apply_overrides(edulint_result + flake8_result + pylint_result, get_overriders())
    result = apply_tweaks(result, get_tweakers(), config)
    return sort(files_or_dirs, result)


def translate(lang_translations: LangTranslations, problem: Problem):
    translation = lang_translations.get(problem.code)
    if translation is None:
        translation = lang_translations.get(problem.symbol)
        if translation is None:
            return problem

    problem.text = translation.translate(problem.code, problem.text)
    return problem


def lint_many(
    partition: List[Tuple[List[str], ImmutableConfig, LangTranslations]]
) -> List[Problem]:
    return [
        translate(lang_translations, problem)
        for files_or_dirs, config, lang_translations in partition
        for problem in lint(files_or_dirs, config)
    ]



import os
from typing import Dict, Any
from collections import defaultdict

import tomli
import tomli_w

from pylint_data import extract_from_pylint
from thonny_data import extract_from_edulint


SCRIPT_PATH = os.path.dirname(os.path.abspath(__file__))
PYLINT_FILE = extract_from_pylint.EDULINT_TOML
THONNY_FILE = extract_from_edulint.EDULINT_TOML
MANUAL_FILE = os.path.join(SCRIPT_PATH, "edulint_manual.toml")
OUTPUT_FILE = os.path.join(SCRIPT_PATH, "../explanations.toml")


def load_toml_file(filename: str) -> Dict[str, Any]:
    with open(filename, "rb") as f:
        return tomli.load(f)


def combine_sources():
    pylint_data = load_toml_file(PYLINT_FILE)
    thonny_data = load_toml_file(THONNY_FILE)
    manual_data = load_toml_file(MANUAL_FILE)

    answer: Dict[str, Dict[str, str]] = defaultdict(dict)

    for data_source in [pylint_data, thonny_data, manual_data]:
        for check_id, check_data in data_source.items():
            for check_field_name, check_field_value in check_data.items():
                if not check_field_value:
                    continue
                answer[check_id][check_field_name] = check_field_value

    with open(OUTPUT_FILE, "wb") as f:
        tomli_w.dump(answer, f, multiline_strings=True)


def main():
    extract_from_pylint.process_from_stored_data()
    extract_from_edulint.process_from_stored_data()
    combine_sources()


if __name__ == "__main__":
    main()



# Copied from https://github.com/GiraffeReversed/edulint-web/blob/main/utils.py

from typing import List, Optional
from dataclasses import dataclass
import functools

from packaging import version as packaging_version


@dataclass
class Version(packaging_version.Version):
    def __init__(self, version: str) -> None:
        super().__init__(version)

    def is_not_full_release(self) -> bool:
        return not (self.is_prerelease or self.is_postrelease or self.is_devrelease)

    def name(self) -> str:
        return str(self).replace(".", "_")

    def dir(self, prefix: str) -> str:
        return f"{prefix}_{self.name()}"

    def __str__(self) -> str:
        return super().__str__()

    def __repr__(self) -> str:
        return super().__repr__()

    @staticmethod
    def parse(version_raw: str) -> Optional["Version"]:
        try:
            return Version(version_raw)
        except packaging_version.InvalidVersion:
            return None


@functools.lru_cache()
def get_available_versions(versions_raw: List[str]) -> List[Version]:
    return [Version(v) for v in versions_raw]



import pytest
from edulint.options import Option
from edulint.config.arg import Arg
from edulint.linting.problem import Problem
from test_utils import lazy_problem, apply_and_lint, create_apply_and_lint
from typing import List


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "def xxx(x):",
        "    while True:",
        "        if x:",
        "            break",
    ], [
        lazy_problem().set_line(2)
        .set_text("The while condition can be replaced with '<negated x>'")
    ]),
])
def test_while_true_break(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=no-while-true")],
        [p.set_code("R6301") for p in expected_output]
    )


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "n = 10",
        "i = 0",
        "while i < n:",
        "    print('foo' + str(i))",
        "    i += 1",
    ], [lazy_problem().set_line(3)]),
])
def test_use_for_loop_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=use-for-loop")],
        [p.set_code("R6305") for p in expected_output]
    )


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "for i in range(10):",
        "    print('foo')",
        "    i += 1"
    ], [
        lazy_problem().set_line(3).set_text("Changing the control variable i of a for loop has no effect.")
    ]),
    ([
        "for _ in range(10):",
        "    _ = 'a,b,c,d'.split(',', 1)",
    ], [
    ]),
    ([
        "for _ in range(10):",
        "    _, rest = 'a,b,c,d'.split(',', 1)",
    ], [
    ]),
    ([
        "for _ in range(10):",
        "    for _ in range(10):"
        "        _, rest = 'a,b,c,d'.split(',', 1)",
    ], [
    ])
])
def test_changing_control_variable_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines,
        [Arg(Option.PYLINT, "--enable=changing-control-variable")],
        [p.set_code("R6304") for p in expected_output]
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("hw14062.py", [
        lazy_problem().set_code("R6302").set_line(29)
        .set_text("Use tighter range boundaries, the first iteration never happens.")
    ]),
    ("hw34666.py", [
        lazy_problem().set_code("R6302").set_line(245)
        .set_text("Use tighter range boundaries, the last iteration never happens.")
    ]),
    ("m1630.py", [
        lazy_problem().set_code("R6302").set_line(107)
        .set_text("Use tighter range boundaries, the last iteration never happens.")
    ]),
])
def test_tighter_bounds_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=use-tighter-boundaries")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("010666-even.py", [lazy_problem().set_line(9)]),
    ("012630-next.py", []),
    ("012986-nested.py", [lazy_problem().set_line(35)]),
    ("013450-next.py", []),
    ("014286-coins.py", []),
    ("014556-prime.py", [lazy_problem().set_line(13)]),
    ("014962-sequence.py", [lazy_problem().set_line(18)]),
    ("015794-amicable.py", [lazy_problem().set_line(14)]),
    ("016119-coins.py", [lazy_problem().set_line(22)]),
    ("017667-coins.py", []),
    ("023140-credit.py", []),
    ("043611-swap_columns.py", []),
    ("05ad83ee13-p5_credit.py", []),
    ("064fe05979.py", []),
    ("11fa2efd2c-exam1.py", []),  # would force using for control after loop
    ("3949eabe39-hw3.py", []),
    ("5ac87a0e60-hw3.py", [lazy_problem().set_line(152), lazy_problem().set_line(164)]),
    ("62398f5c25.py", [lazy_problem().set_line(111)]),
    ("b2c113956e-d_mancala.py", [lazy_problem().set_line(31)]),
    ("f56a8957e2-p3_triples.py", []),
    ("uc_1597_09_16.py", []),
    ("uc_3776_30_15.py", []),
    ("uc_7004_17_14.py", [lazy_problem().set_line(3)]),
])
def test_use_for_loop_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=use-for-loop")],
        [problem.set_code("R6305") for problem in expected_output]
    )


# OBJ_DEF = [
#     "class Obj:",
#     "    def __init__(self):",
#     "        self.elems = []",
# ]

# @pytest.mark.parametrize("lines,expected_output", [
#     (OBJ_DEF + [
#         "obj = Obj()",
#         "for v in obj.elems:",
#         "    obj.elems.append('x')",
#     ], [lazy_problem().set_line(6)]),
#     (OBJ_DEF + [
#         "obj = Obj()",
#         "for v in obj.elems:",
#         "    obj = Obj()",
#         "    obj.elems.append('y')"
#     ], []),
#     (OBJ_DEF + [
#         "obj = Obj()",
#         "for v in obj.elems:",
#         "    obj.elems = []",
#     ], [lazy_problem().set_line(6)]),
#     (OBJ_DEF + [
#         "obj = Obj()",
#         "for v in obj.elems:",
#         "    obj.elems.append('x')",
#         "    obj = Obj()",
#         "    obj.elems.append('y')"
#     ], [lazy_problem().set_line(6)]),
# ])
# def test_modifying_iterated_custom(lines: List[str], expected_output: List[Problem]) -> None:
#     create_apply_and_lint(
#         lines,
#         [Arg(Option.PYLINT, "--enable=modifying-iterated-structure")],
#         [problem.set_code("R6303") for problem in expected_output]
#     )

@pytest.mark.parametrize("filename,expected_output", [
    ("034636-right_cellular.py", []),
    ("104526-subset_sum.py", [
        lazy_problem().set_line(17)
        .set_text("Iterated structure nums is being modified inside the for loop body. "
                  "Use while loop or iterate over a copy."),
        lazy_problem().set_line(22)
    ]),
    ("035794-partition.py", []),
    ("2902f19359.py", []),
    ("eba554ae42-HW4.py", [lazy_problem().set_line(189)]),
    ("hw24328.py", [
        lazy_problem().set_line(52)
        .set_text("Iterated structure row is being modified inside the for loop body. "
                  "Use while loop or iterate over a copy."),
        lazy_problem().set_line(53),
        lazy_problem().set_line(56),
        lazy_problem().set_line(99),
        lazy_problem().set_line(100),
        lazy_problem().set_line(103),
    ]),
    ("hw44081.py", [
        lazy_problem().set_line(149)
        .set_text("Iterated structure traders is being modified inside the for loop body. "
                  "Use while loop or iterate over a copy.")
    ]),
])
def test_modifying_iterated_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=modifying-iterated-structure")],
        [problem.set_code("R6303") for problem in expected_output]
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("hw33176.py", []),
    ("m4521.py", [
        lazy_problem().set_line(117).set_text("Inner for loop shadows outer for loop's control variable i.")
    ])
])
def test_shadowed_control_var_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=loop-shadows-control-variable")],
        expected_output
    )


@pytest.mark.parametrize("filename,expected_output", [
    ("014659-next.py", [
        lazy_problem().set_line(24).set_text("Changing the control variable i of a for loop has no effect.")
    ]),
    ("hw24226.py", [
        lazy_problem().set_line(77).set_text("Changing the control variable j of a for loop has no effect."),
        lazy_problem().set_line(101).set_text("Changing the control variable p of a for loop has no effect."),
    ]),
    ("hw24387.py", [
        lazy_problem().set_line(98).set_text("Changing the control variable i of a for loop has no effect."),
        lazy_problem().set_line(129),
    ]),
    ("hw34406.py", []),
    ("hw34666.py", [
        lazy_problem().set_line(40),
        lazy_problem().set_line(43),
        lazy_problem().set_line(51),
        lazy_problem().set_line(54),
        lazy_problem().set_line(57),
        lazy_problem().set_line(71),
        lazy_problem().set_line(74),
        lazy_problem().set_line(77),
    ]),
])
def test_changing_control_var_files(filename: str, expected_output: List[Problem]) -> None:
    apply_and_lint(
        filename,
        [Arg(Option.PYLINT, "--enable=changing-control-variable")],
        [problem.set_code("R6304") for problem in expected_output]
    )


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "A = list(range(10))",
        "B = []",
        "for x in range(len(A)):",
        "    for y in range(len(A)):",
        "        B.append(A[x])",
        "        B.append(y)",
    ], [
        lazy_problem().set_line(3)
        .set_text('Iterate directly: "for var in A" (with appropriate name for "var")'),
    ]),
    ([
        "A = list(range(10))",
        "for i in range(5, len(A)):",
        "    print(A[i])"
    ], []),
    ([
        "A = list(range(10))",
        "a = 5",
        "for i in range(a, len(A)):",
        "    print(A[i])"
    ], []),
    ([
        "A = list(range(10))",
        "a = 5",
        "for i in range(a, len(A)):",
        "    print(i, A[i])",
    ], []),
    ([
        "A = list(range(10))",
        "B = []",
        "for x in range(len(A)):",
        "    x += 1",
        "    B.append(A[x])",
        "    B.append(A[x + 1])",
    ], []),
    ([
        "A = list(range(10))",
        "B = []",
        "for x in range(len(A)):",
        "    B.append(A[x + 1])",
    ], []),
    ([
        "A = list(range(10))",
        "for x in range(len(A)):",
        "    A[x] = A[x] + 1"
    ], []),
    ([
        "A = list(range(10))",
        "B = []",
        "for x in range(len(A)):",
        "    for x in range(len(B)):",
        "        B.append(A[x])",
    ], []),
])
def test_use_foreach_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines, [Arg(Option.PYLINT, "--enable=use-foreach")], expected_output
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("00bcea235f.py", [lazy_problem().set_line(23)]),
    ("014771-p2_nested.py", [
            lazy_problem().set_line(25)
            .set_text('Iterate directly: "for var in A" (with appropriate name for "var")'),
            lazy_problem().set_line(35)
            .set_text('Iterate directly: "for var in A" (with appropriate name for "var")'),
    ]),
    ("015080-p4_geometry.py", []),
    ("01c9f86c74-task_2.py", []),
    ("03-d4_points.py", []),
    ("045294-p4_vigenere.py", []),
    ("046542-polybius.py", [lazy_problem().set_line(67)]),
    ("05784ba6f8.py", [lazy_problem().set_line(46)]),
    ("105119-p5_template.py", []),
    ("35a73a1aa0.py", []),
    ("5e3d9ea337-greater.py", []),
    ("b313a82293.py", [
        lazy_problem().set_line(25),
        lazy_problem().set_line(26),
        lazy_problem().set_line(33),
    ]),
    ("be217c2c64-midterm_alt.py", []),
    ("d9f34aa130-fixpoint.py", [lazy_problem().set_line(31)]),
    ("umime_count_a.py", [
        lazy_problem().set_line(3)
        .set_text('Iterate directly: "for var in text" (with appropriate name for "var")'),
    ]),
])
def test_use_foreach(filename: str, expected_output: List[Problem]):
    apply_and_lint(filename, [Arg(Option.PYLINT, "--enable=use-foreach")], expected_output)


@pytest.mark.parametrize("lines,expected_output", [
    ([
        "A = list(range(10))",
        "B = []",
        "for x in range(len(A)):",
        "    for y in range(len(A)):",
        "        B.append(A[x])",
        "        B.append(y)",
    ], []),
    ([
        "A = list(range(10))",
        "for i in range(5, len(A)):",
        "    print(A[i])"
    ], []),
    ([
        "A = list(range(10))",
        "a = 5",
        "for i in range(a, len(A)):",
        "    print(A[i])"
    ], []),
    ([
        "A = list(range(10))",
        "a = 5",
        "for i in range(a, len(A)):",
        "    print(i, A[i])",
    ], []),
    ([
        "A = list(range(10))",
        "B = []",
        "for x in range(len(A)):",
        "    x += 1",
        "    B.append(A[x])",
        "    B.append(A[x + 1])",
    ], []),
    ([
        "A = list(range(10))",
        "B = []",
        "for x in range(len(A)):",
        "    B.append(A[x + 1])",
    ], []),
    ([
        "A = list(range(10))",
        "for x in range(len(A)):",
        "    A[x] = A[x] + 1"
    ], [  # unnecessary?
            # lazy_problem()
            # .set_line(2)
            # .set_text(
            #     'Iterate using enumerate: "for x, var in enumerate(A)" (with appropriate name for "var")'
            # ),
    ]),
    ([
        "A = list(range(10))",
        "for x in range(len(A)):",
        "    print(x, A[x])"
    ], [
        lazy_problem().set_line(2)
        .set_text('Iterate using enumerate: "for x, var in enumerate(A)" (with appropriate name for "var")'),
    ]),
    ([
        "A = list(range(10))",
        "B = []",
        "for x in range(len(A)):",
        "    for x in range(len(B)):",
        "        B.append(A[x])",
    ], [])
])
def test_use_enumerate_custom(lines: List[str], expected_output: List[Problem]) -> None:
    create_apply_and_lint(
        lines, [Arg(Option.PYLINT, "--enable=use-enumerate")], expected_output
    )

@pytest.mark.parametrize("filename,expected_output", [
    ("014771-p2_nested.py", []),
    ("015080-p4_geometry.py", []),
    ("01c9f86c74-task_2.py", []),
    ("03-d4_points.py", []),
    ("041378925b-f_tetris.py", [lazy_problem().set_line(134), lazy_problem().set_line(135)]),
    ("045294-p4_vigenere.py", []),
    ("046542-polybius.py", [
        lazy_problem().set_line(44),
        lazy_problem().set_line(45),
        lazy_problem().set_line(46),  # multi-step
        lazy_problem().set_line(68),
        lazy_problem().set_line(69),  # multi-step
    ]),
    ("105119-p5_template.py", []),
    ("35a73a1aa0.py", []),
    ("7097dd29de-p1_lists.py", [lazy_problem().set_line(46)]),
    ("be217c2c64-midterm_alt.py", []),
    ("d9f34aa130-fixpoint.py", []),
    ("umime_count_a.py", []),
])
def test_use_enumerate(filename: str, expected_output: List[Problem]):
    apply_and_lint(filename, [Arg(Option.PYLINT, "--enable=use-enumerate")], expected_output)



from astroid import nodes, Context  # type: ignore
from functools import reduce
from typing import TYPE_CHECKING, List, Iterator, Tuple, Optional

from pylint.checkers import BaseChecker  # type: ignore
from pylint.checkers.utils import only_required_for_messages

if TYPE_CHECKING:
    from pylint.lint import PyLinter  # type: ignore

from edulint.linting.checkers.utils import (
    get_range_params,
    get_const_value,
    is_builtin,
    requires_data_dependency_analysis,
)
from edulint.linting.analyses.cfg.utils import successors_from_loc, get_cfg_loc, CFGLoc
from edulint.linting.analyses.var_events import VarEventType, Variable, VarEvent, strip_to_name
from edulint.linting.analyses.data_dependency import (
    vars_in,
    node_to_var,
    MODIFYING_EVENTS,
    get_events_for,
    get_events_by_var,
    filter_events_for,
)


class UnsuitedLoop(BaseChecker):
    name = "unsuited-loop"
    msgs = {
        "R6301": (
            "The while condition can be replaced with '<negated %s>'",
            "no-while-true",
            "Emitted when the condition of a while loop is 'True' unnecessarily.",
        ),
        "R6302": (
            "Use tighter range boundaries, the %s iteration never happens.",
            "use-tighter-boundaries",
            "Emitted when the boundaries of a range can be made tighter.",
        ),
        "R6303": (
            "Iterated structure %s is being modified inside the for loop body. Use while loop or iterate over a copy.",
            "modifying-iterated-structure",
            "Emitted when the structure that is being iterated over is being modified in the for loops body.",
        ),
        "R6304": (
            "Changing the control variable %s of a for loop has no effect.",
            "changing-control-variable",
            "Emitted when the control variable of a for loop is being changed.",
        ),
        "R6305": (
            "You should use for loop instead of this form of while loop.",
            "use-for-loop",
            "Emitted when a while loop can be transformed into a for loop.",
        ),
        "R6306": (
            "Inner for loop shadows outer for loop's control variable %s.",
            "loop-shadows-control-variable",
            "Emitted when a for loop shadows control variable of an outer for loop.",
        ),
        "R6307": (
            'Iterate directly: "for var in %s" (with appropriate name for "var")',
            "use-foreach",
            "Emitted when a for-range loop is used while a for-each loop would suffice.",
            {"old_names": [("R6101", "old-use-foreach")]},
        ),
        "R6308": (
            'Iterate using enumerate: "for %s, var in enumerate(%s)" (with appropriate name for "var")',
            "use-enumerate",
            "Emitted when a for-range loop is used with the element at each index is accessed as well.",
            {"old_names": [("R6102", "old-use-enumerate")]},
        ),
    }

    def _check_no_while_true(self, node: nodes.While) -> None:
        if not isinstance(node.test, nodes.Const) or not node.test.bool_value():
            return

        first = node.body[0]
        if isinstance(first, nodes.If) and isinstance(first.body[-1], nodes.Break):
            self.add_message("no-while-true", node=node, args=(first.test.as_string(),))

    @requires_data_dependency_analysis
    def _check_use_for_loop(self, node: nodes.While) -> None:
        # TODO allow different increments?
        def get_change_value(node: nodes.NodeNG) -> Optional[int]:
            if not isinstance(node, nodes.AssignName):
                return None

            expr = node.parent
            const = None
            op = None
            if isinstance(expr, nodes.AugAssign) and expr.op in ("+=", "-="):
                const = get_const_value(expr.value)
                op = expr.op[0]
            if isinstance(expr, nodes.Assign) and isinstance(expr.value, nodes.BinOp):
                binop = expr.value
                if binop.op in ("+", "-"):
                    op = binop.op
                    if binop.left.as_string() == node.as_string():
                        const = get_const_value(binop.right)
                    if binop.right.as_string() == node.as_string():
                        const = get_const_value(binop.left)
            if const is None:
                return None

            return const if op == "+" else -const

        def var_can_be_from_range(var: Variable, test: nodes.NodeNG) -> bool:
            events = list(get_events_for([var], [test]))
            return len(events) == 1 and events[0].node.parent == test

        def stop_on_event_node_wrap(events):
            event_nodes = {get_cfg_loc(e.node).node for e in events}

            def stop_on(loc: CFGLoc):
                return (
                    loc is None
                    or isinstance(loc.node, (nodes.Break, nodes.Return))
                    or loc.node in event_nodes
                )

            return stop_on

        def event_on_every_path(node: nodes.While, events):
            if node.next_sibling() is None:
                node.cfg_loc.block.locs.append(None)

            result = True
            for loc in successors_from_loc(
                node.body[0].cfg_loc,
                stop_on_loc=stop_on_event_node_wrap(events),
                include_start=True,
                include_end=True,
                explore_functions=True,
                explore_classes=True,
            ):
                if (
                    loc is None
                    or node not in loc.node.node_ancestors()
                    or (
                        isinstance(loc.node, (nodes.Break, nodes.Return))
                        and any(
                            node not in use.node.node_ancestors()
                            for event in events
                            for use in event.uses
                        )
                    )
                ):
                    result = False
                    break
            if node.cfg_loc.block.locs[-1] is None:
                node.cfg_loc.block.locs.pop()

            return result

        def event_in_loop(parent_loop: nodes.While, event):
            parent = event.node
            while parent != parent_loop:
                if isinstance(parent, (nodes.For, nodes.While)):
                    return True
                parent = parent.parent
            return False

        def modification_after_modification(events):
            return any(e.uses != events[0].uses for e in events)

        def changes_towards_limit(test, only_var, change):
            lt, (op, rt) = test.left, test.ops[0]
            if only_var in vars_in(lt):
                return change > 0 and op in ("<", "<=") or change < 0 and op in (">", ">=")
            assert only_var in vars_in(rt)
            return change < 0 and op in ("<", "<=") or change > 0 and op in (">", ">=")

        test = node.test
        if (
            not isinstance(test, nodes.Compare)
            or len(test.ops) != 1
            or test.ops[0][0] not in ("<", "<=", ">", ">=")
        ):
            return

        test_vars = list(vars_in(test).keys())
        if len(test_vars) == 0:
            return

        events_by_var = get_events_by_var(test_vars, node.body, MODIFYING_EVENTS)

        if len(events_by_var) != 1:
            return

        only_var, events = next(iter(events_by_var.items()))
        change = reduce(
            lambda r, c: c if c is not None and r is not None and c == r else None,
            [get_change_value(e.node) for e in events],
        )
        if change not in (1, -1):
            return

        if (
            var_can_be_from_range(only_var, test)
            and changes_towards_limit(test, only_var, change)
            and not modification_after_modification(events)
            and not any(event_in_loop(node, e) for e in events)
            and event_on_every_path(node, events)
        ):
            self.add_message("use-for-loop", node=node)

    @only_required_for_messages("no-while-true", "use-for-loop")
    def visit_while(self, node: nodes.While) -> None:
        self._check_no_while_true(node)
        self._check_use_for_loop(node)

    def _check_use_tighter_bounds(self, node: nodes.For) -> None:
        def compares_equality(node: nodes.NodeNG) -> bool:
            return isinstance(node, nodes.Compare) and len(node.ops) == 1 and node.ops[0][0] == "=="

        def compares_to_start(node: nodes.Compare, var: str, start: nodes.NodeNG) -> bool:
            left, (_, right) = node.left, node.ops[0]
            left, right = left.as_string(), right.as_string()
            return (left == var and right == start.as_string()) or (
                left == start.as_string() and right == var
            )

        def is_last_before_end(node: nodes.NodeNG, stop: nodes.NodeNG, step: nodes.NodeNG) -> bool:
            const_node = get_const_value(node)
            const_stop = get_const_value(stop)
            const_step = get_const_value(step)
            return (
                isinstance(node, nodes.BinOp)
                and node.op == "-"
                and node.left.as_string() == stop.as_string()
                and node.right.as_string() == step.as_string()
            ) or (
                isinstance(const_node, int)
                and isinstance(const_stop, int)
                and isinstance(const_step, int)
                and const_node == const_stop - const_step
            )

        def compares_to_last_before_end(
            node: nodes.Compare, var: str, stop: nodes.NodeNG, step: nodes.NodeNG
        ) -> bool:
            left, (_, right) = node.left, node.ops[0]
            return (left.as_string() == var and is_last_before_end(right, stop, step)) or (
                is_last_before_end(left, stop, step) and right.as_string() == var
            )

        def relevant_nodes(body: List[nodes.NodeNG]) -> Iterator[nodes.If]:
            first = body[0]
            if isinstance(first, nodes.If):
                yield first
                while first.has_elif_block():
                    first = first.orelse[0]
                    yield first

            if len(body) > 1 and isinstance(body[-1], nodes.If):
                yield body[-1]

        range_params = get_range_params(node.iter)
        if range_params is None:
            return

        start, stop, step = range_params

        var = node.target.as_string()
        if isinstance(node.body[0], nodes.If):
            if_ = node.body[0]
            test = if_.test

            if (
                compares_equality(test)
                and len(if_.body) == 1
                and type(if_.body[-1]) in (nodes.Break, nodes.Continue)
            ):
                if compares_to_start(test, var, start):
                    self.add_message("use-tighter-boundaries", node=node, args=("first",))
                elif compares_to_last_before_end(test, var, stop, step):
                    self.add_message("use-tighter-boundaries", node=node, args=("last",))
            # TODO expand, tips: len(node.body) == 1, isinstance(if_.body[-1], nodes.Return), allow nodes in the middle,
            #                    allow longer bodies

    @staticmethod
    def _get_block_line(node: nodes.NodeNG) -> nodes.NodeNG:
        while not node.is_statement:
            node = node.parent
        return node

    @staticmethod
    def _get_last_block_line(node: nodes.NodeNG) -> nodes.NodeNG:
        node = UnsuitedLoop._get_block_line(node)

        while node.next_sibling() is not None:
            node = node.next_sibling()
        return node

    @requires_data_dependency_analysis
    def _check_modifying_iterable(self, node: nodes.For) -> None:
        iterated = node.iter
        if not isinstance(iterated, nodes.Name):  # TODO allow any node type
            return

        iterated_var = node_to_var(iterated)
        if iterated_var is None:
            return

        # TODO reassigning iterated structure message?
        for event in get_events_for([iterated_var], node.body, MODIFYING_EVENTS):
            if (
                event.type in (VarEventType.ASSIGN, VarEventType.REASSIGN)
                and get_cfg_loc(event.node).node.parent == node
            ):
                break
            if (
                event.type == VarEventType.MODIFY
                and event.is_direct_modify()
                and not isinstance(
                    self._get_last_block_line(event.node), (nodes.Break, nodes.Return)
                )
            ):
                self.add_message(
                    "modifying-iterated-structure", node=event.node.parent, args=iterated_var.name
                )

    @requires_data_dependency_analysis
    def _check_control_variable_changes(self, node: nodes.For) -> None:
        range_params = get_range_params(node.iter)
        if range_params is None or not isinstance(node.target, nodes.AssignName):
            return

        control_var = node_to_var(node.target)
        assert control_var is not None
        if control_var.name.startswith("_"):
            return

        for event in get_events_for([control_var], node.body, MODIFYING_EVENTS):
            if isinstance(event.node.parent, nodes.For) and event.node == event.node.parent.target:
                self.add_message(
                    "loop-shadows-control-variable", node=event.node, args=control_var.name
                )
                continue

            if len(event.uses) == 0:
                self.add_message(
                    "changing-control-variable", node=event.node.parent, args=(control_var.name)
                )

    @requires_data_dependency_analysis
    def _check_improve_for_loop(self, node: nodes.For) -> None:
        def is_for_range_len(node: nodes.For) -> Tuple[bool, nodes.NodeNG]:
            range_params = get_range_params(node.iter)
            if range_params is None:
                return False, None

            start, stop, step = range_params
            return (
                isinstance(start, nodes.Const)
                and start.value == 0
                and isinstance(stop, nodes.Call)
                and is_builtin(stop.func, "len")
                and len(stop.args) == 1
                and isinstance(step, nodes.Const)
                and step.value == 1
            ), stop

        def get_indexing_events(structure_events: List[VarEvent]) -> List[VarEvent]:
            return [
                e
                for e in structure_events
                if isinstance(e.node.parent, nodes.Subscript) and e.node == e.node.parent.value
            ]

        def is_structure_indexed_just_by(
            index_var: Variable, indexing_events: List[VarEvent]
        ) -> bool:
            for event in indexing_events:
                slice_ = event.node.parent.slice
                if not isinstance(slice_, nodes.Name) or slice_.name != index_var.name:
                    return False
            return True

        def is_structure_assigned_into(structure_events: List[VarEvent]) -> bool:
            return any(
                e.type != VarEventType.READ
                and not isinstance(
                    e.node.parent.parent, (nodes.Subscript, nodes.Attribute, nodes.Call)
                )
                for e in structure_events
            )

        def is_structure_directly_modified(structure_events: List[VarEvent]) -> bool:
            return any(
                (
                    (e.type == VarEventType.MODIFY and e.is_direct_modify())
                    or (e.type not in (VarEventType.READ, VarEventType.MODIFY))
                )
                and not isinstance(self._get_last_block_line(e.node), (nodes.Break, nodes.Return))
                for e in structure_events
            )

        def is_structure_read(indexing_events: List[VarEvent]) -> bool:
            return any(e.node.parent.ctx == Context.Load for e in indexing_events)

        def is_index_modified(index_events: List[VarEvent]) -> bool:
            return any(e.type != VarEventType.READ for e in index_events)

        def is_index_used_just_for(structure: nodes.NodeNG, index_events: List[VarEvent]) -> bool:
            for event in index_events:
                if (
                    isinstance(event.node.parent, nodes.Subscript)
                    and event.node == event.node.parent.slice
                    # TODO use vars_in for next test
                    and event.node.parent.value.as_string() != structure.as_string()
                ):
                    return False
            return True

        def is_index_used_outside_subscript(
            index_events: List[VarEvent], structure: nodes.NodeNG
        ) -> bool:
            for event in index_events:
                if (
                    not isinstance(event.node.parent, nodes.Subscript)
                    or event.node != event.node.parent.slice
                    or event.node.parent.value.as_string() != structure.as_string()
                ):
                    return True
            return False

        is_, stop = is_for_range_len(node)
        if not is_:
            return

        structure = stop.args[0]
        if not isinstance(structure, (nodes.Name, nodes.Attribute, nodes.Subscript)):
            return

        stripped = strip_to_name(structure)
        if stripped is None:
            return

        structure_var = node_to_var(stripped)
        index_var = node_to_var(node.target)
        if structure_var is None or index_var is None:
            return

        events_by_var = get_events_by_var([structure_var, index_var], node.body)
        events_by_var[structure_var] = filter_events_for(structure, events_by_var[structure_var])

        indexing_events = get_indexing_events(events_by_var[structure_var])
        if (
            not is_structure_read(indexing_events)
            or not is_structure_indexed_just_by(index_var, indexing_events)
            or not is_index_used_just_for(structure, events_by_var[index_var])
            or is_index_modified(events_by_var[index_var])
            or is_structure_directly_modified(events_by_var[structure_var])
            or is_structure_assigned_into(events_by_var[structure_var])
        ):
            return

        if is_index_used_outside_subscript(events_by_var[index_var], structure):
            self.add_message("use-enumerate", args=(index_var.name, structure_var.name), node=node)
        else:
            self.add_message("use-foreach", args=structure_var.name, node=node)

    @only_required_for_messages(
        "use-tighter-boundaries",
        "modifying-iterated-structure",
        "changing-control-variable",
        "loop-shadows-control-variable",
        "use-enumerate",
        "use-foreach",
    )
    def visit_for(self, node: nodes.For) -> None:
        self._check_use_tighter_bounds(node)
        self._check_modifying_iterable(node)
        self._check_control_variable_changes(node)
        self._check_improve_for_loop(node)


def register(linter: "PyLinter") -> None:
    """This required method auto registers the checker during initialization.
    :param linter: The linter to register the checker to.
    """
    linter.register_checker(UnsuitedLoop(linter))
